"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from abc import ABC, abstractmethod
from typing import Callable

import numpy as np
from sklearn.base import BaseEstimator

from mlrose_ky.neural.fitness.network_weights import NetworkWeights
from mlrose_ky.neural.utils import unflatten_weights
from mlrose_ky.opt_probs import ContinuousOpt


class _NNBase(BaseEstimator, ABC):
    """
    Abstract base class for neural network models.

    Defines the necessary methods and utility functions for training and
    predicting with neural network models.
    """

    @abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray = None, init_weights: np.ndarray = None):
        """Fit the neural network to the data."""
        raise NotImplementedError("Subclasses must implement fit method")

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Use the model to predict data labels for a given feature array."""
        raise NotImplementedError("Subclasses must implement predict method")

    @staticmethod
    def _calculate_state_size(node_list: list[int]) -> int:
        """
        Calculate the total number of nodes in the network layers.

        Parameters
        ----------
        node_list : list of int
            List of the number of nodes in each layer, including input
            and output layers.

        Returns
        -------
        int
            The total number of nodes in the network.
        """
        return sum(node_list[i] * node_list[i + 1] for i in range(len(node_list) - 1))

    @staticmethod
    def _build_node_list(X: np.ndarray, y: np.ndarray, hidden_nodes: list[int], bias: bool = False) -> list[int]:
        """
        Build a list of nodes in each layer of the network.

        Parameters
        ----------
        X : np.ndarray
            Feature dataset with each row representing a single observation.
        y : np.ndarray
            Array of data labels.
        hidden_nodes : list of int
            List of the number of nodes in the hidden layers.
        bias : bool, optional, default=False
            Whether to include a bias term in the network.

        Returns
        -------
        list of int
            A list containing the number of nodes in each layer of the network.
        """
        input_nodes = np.shape(X)[1] + bias
        output_nodes = np.shape(y)[1]

        return [input_nodes] + hidden_nodes + [output_nodes]

    @staticmethod
    def _format_x_y_data(X: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """
        Ensure that the X and y data are correctly formatted.

        Parameters
        ----------
        X : np.ndarray
            Feature dataset with each row representing a single observation.
        y : np.ndarray
            Data labels.

        Returns
        -------
        X : np.ndarray
            Formatted feature dataset.
        y : np.ndarray
            Formatted data labels.

        Raises
        ------
        ValueError
            If the lengths of X and y do not match.
        """
        y = np.array(y)

        if len(np.shape(y)) == 1:
            y = np.reshape(y, [len(y), 1])

        if np.shape(X)[0] != np.shape(y)[0]:
            raise ValueError(f"The length of X ({np.shape(X)[0]}) and y ({np.shape(y)[0]}) must be equal.")

        return X, y

    @staticmethod
    def _build_problem_and_fitness_function(
        X: np.ndarray,
        y: np.ndarray,
        node_list: list[int],
        activation: Callable,
        learning_rate: float,
        clip_max: float,
        bias: bool = False,
        is_classifier: bool = True,
    ) -> tuple[NetworkWeights, ContinuousOpt]:
        """
        Initialize the optimization problem and fitness function.

        Parameters
        ----------
        X : np.ndarray
            Feature dataset.
        y : np.ndarray
            Data labels.
        node_list : list of int
            List of the number of nodes in each layer.
        activation : Callable
            Activation function for the network.
        learning_rate : float
            Learning rate for weight updates.
        clip_max : float
            Maximum value for weights.
        bias : bool, optional, default=False
            Whether to include a bias term in the network.
        is_classifier : bool, optional, default=True
            Whether the network is a classifier.

        Returns
        -------
        NetworkWeights
            The fitness function for optimizing the network weights.
        ContinuousOpt
            The continuous optimization problem for gradient descent.
        """
        fitness = NetworkWeights(X, y, node_list, activation, bias, is_classifier, learning_rate=learning_rate)
        num_nodes = _NNBase._calculate_state_size(node_list)

        problem = ContinuousOpt(
            length=num_nodes, fitness_fn=fitness, maximize=False, min_val=-clip_max, max_val=clip_max, step=learning_rate
        )

        return fitness, problem

    @staticmethod
    def _predict(
        X: np.ndarray,
        fitted_weights: np.ndarray,
        node_list: list[int],
        input_activation: Callable,
        output_activation: Callable,
        bias: bool = False,
        is_classifier: bool = True,
    ) -> tuple[np.ndarray, np.ndarray | None]:
        """
        Predict data labels based on the fitted weights of the network.

        Parameters
        ----------
        X : np.ndarray
            Feature dataset.
        fitted_weights : np.ndarray
            Flattened weight array for the network.
        node_list : list of int
            List of the number of nodes in each layer.
        input_activation : Callable
            Activation function for the hidden layers.
        output_activation : Callable
            Activation function for the output layer.
        bias : bool, optional, default=False
            Whether to include a bias term in the network.
        is_classifier : bool, optional, default=True
            Whether the network is a classifier.

        Returns
        -------
        y_pred : np.ndarray
            Predicted labels for the input dataset.
        predicted_probs : np.ndarray or None
            Predicted probabilities for the input dataset, if the network is a classifier.
        """
        if not node_list:
            raise ValueError("node_list cannot be empty.")

        weights = list(unflatten_weights(fitted_weights, node_list))

        if bias:
            ones = np.ones([np.shape(X)[0], 1])
            inputs = np.hstack((X, ones))
        else:
            inputs = X

        y_pred = np.empty(0)  # Initialize y_pred to prevent uninitialized warning
        predicted_probs = None  # Initialize predicted_probs

        for i in range(len(weights)):
            outputs = np.dot(inputs, weights[i])

            if i < len(weights) - 1:
                inputs = input_activation(outputs)
            else:
                y_pred = output_activation(outputs)

        if is_classifier:
            predicted_probs = y_pred

            if node_list[-1] == 1:
                y_pred = np.round(y_pred).astype(int)
            else:
                zeros = np.zeros_like(y_pred)
                zeros[np.arange(len(y_pred)), np.argmax(y_pred, axis=1)] = 1
                y_pred = zeros.astype(int)

        return y_pred, predicted_probs



"""Classes for defining optimization problem objects for Simulated Annealing (SA)."""

# Author: Genevieve Hayes
# License: BSD 3-clause

from .arith_decay import ArithDecay
from .custom_schedule import CustomSchedule
from .exp_decay import ExpDecay
from .geom_decay import GeomDecay



"""Neural network 'Sigmoid' activation function."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

import warnings

import numpy as np

from mlrose_ky.decorators import short_name

warnings.filterwarnings("ignore")


@short_name("sigmoid")
def sigmoid(x: np.ndarray, deriv: bool = False) -> np.ndarray:
    """Sigmoid activation function.

    Parameters
    ----------
    x: np.ndarray
        Array containing input data.

    deriv: bool, optional, default=False
        Whether to return the function (when False) or its derivative (when True).

    Returns
    -------
    fx: np.ndarray
        Value of activation function at x.
    """
    fx = 1 / (1 + np.exp(-x))

    if deriv:
        fx *= 1 - fx

    return fx



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np


class _OptProb:
    """Base class for optimization problems.

    Parameters
    ----------
    length : int
        Number of elements in the state vector.
    fitness_fn : Any
        Object to implement the fitness function for optimization.
    maximize : bool, default=True
        Whether to maximize the fitness function.
        Set :code:`False` for minimization problem.

    Attributes
    ----------
    length : int
        Length of the state vector.
    fitness_fn : Any
        Fitness function for the optimization problem.
    maximize : float
        Maximization factor, 1.0 for maximization, -1.0 for minimization.
    state : np.ndarray
        Current state vector.
    neighbors : np.ndarray
        Array containing neighboring states.
    fitness : float
        Fitness value of the current state.
    population : np.ndarray
        Array containing the current population.
    pop_fitness : np.ndarray
        Array containing the fitness values for the current population.
    mate_probs : np.ndarray
        Array containing the mate probabilities for the current population.
    fevals : dict
        Dictionary for tracking function evaluations.
    fitness_evaluations : int
        Counter for the number of fitness evaluations.
    current_iteration : int
        Current iteration number in the optimization process.
    """

    def __init__(self, length: int, fitness_fn: Any, maximize: bool = True):
        if length < 0:
            raise ValueError("length must be a positive integer.")
        elif not isinstance(length, int):
            if length.is_integer():
                self.length: int = int(length)
            else:
                raise ValueError("length must be a positive integer.")
        else:
            self.length: int = length

        self.state: np.ndarray = np.array([0] * self.length)
        self.neighbors: np.ndarray = np.array([])
        self.fitness_fn: Any = fitness_fn
        self.fitness: float = 0.0
        self.population: np.ndarray = np.array([])
        self.pop_fitness: np.ndarray = np.array([])
        self.mate_probs: np.ndarray = np.array([])
        self.fevals: dict = {}
        self.fitness_evaluations: int = 0
        self.current_iteration: int = 0
        self.maximize: float = 1.0 if maximize else -1.0

    def best_child(self) -> np.ndarray:
        """Return the best state in the current population.

        Returns
        -------
        np.ndarray
            State vector defining the best child.
        """
        return self.population[np.argmax(self.pop_fitness)]

    def best_neighbor(self) -> np.ndarray:
        """Return the best neighbor of the current state.

        Returns
        -------
        np.ndarray
            State vector defining the best neighbor.
        """
        fitness_list = [self.eval_fitness(neigh) for neigh in self.neighbors]
        return self.neighbors[np.argmax(fitness_list)]

    def eval_fitness(self, state: np.ndarray) -> float:
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state : np.ndarray
            State vector for evaluation.

        Returns
        -------
        float
            Value of the fitness function.
        """
        if len(state) != self.length:
            raise ValueError(f"State length {len(state)} must match problem length {self.length}")

        fitness = self.maximize * self.fitness_fn.evaluate(state)
        self.fitness_evaluations += 1
        return fitness

    def eval_mate_probs(self) -> None:
        """Calculate the probability of each member of the population reproducing."""
        pop_fitness = np.copy(self.pop_fitness)

        # Set -1*inf values to 0 to avoid dividing by sum of infinity.
        # This forces mate_probs for these pop members to 0.
        pop_fitness[pop_fitness == -1.0 * np.inf] = 0

        # Account for maximize = False
        if self.maximize == -1.0:
            pop_fitness -= np.min(pop_fitness)

        if np.sum(pop_fitness) == 0:
            self.mate_probs = np.ones(len(pop_fitness)) / len(pop_fitness)
        else:
            self.mate_probs = pop_fitness / np.sum(pop_fitness)

    def get_fitness(self) -> float:
        """Return the fitness of the current state vector.

        Returns
        -------
        float
            Fitness value of the current state vector.
        """
        return self.fitness

    def get_adjusted_fitness(self) -> float:
        """Return maximization factor * fitness of the current state vector.

        Returns
        -------
        float
            Fitness value of the current state vector adjusted by the maximization factor.
        """
        return self.maximize * self.fitness

    def get_length(self) -> int:
        """Return the state vector length.

        Returns
        -------
        int
            Length of the state vector.
        """
        return self.length

    def get_mate_probs(self) -> np.ndarray:
        """Return the population mate probabilities.

        Returns
        -------
        np.ndarray
            Numpy array containing mate probabilities of the current population.
        """
        return self.mate_probs

    def get_maximize(self) -> float:
        """Return the maximization multiplier.

        Returns
        -------
        float
            Maximization multiplier.
        """
        return self.maximize

    def get_pop_fitness(self) -> np.ndarray:
        """Return the current population fitness array.

        Returns
        -------
        np.ndarray
            Numpy array containing the fitness values for the current population.
        """
        return self.pop_fitness

    def get_population(self) -> np.ndarray:
        """Return the current population.

        Returns
        -------
        np.ndarray
            Numpy array containing the current population.
        """
        return self.population

    def get_state(self) -> np.ndarray:
        """Return the current state vector.

        Returns
        -------
        np.ndarray
            Current state vector.
        """
        return self.state

    def set_population(self, new_population: np.ndarray) -> None:
        """Set a new population and evaluate its fitness.

        Parameters
        ----------
        new_population : np.ndarray
            Numpy array containing the new population.
        """
        self.population = new_population
        self.evaluate_population_fitness()

    def evaluate_population_fitness(self) -> None:
        """Evaluate the fitness of the current population."""
        self.pop_fitness = np.array([self.eval_fitness(indiv) for indiv in self.population])

    def set_state(self, new_state: np.ndarray) -> None:
        """Set a new state vector and evaluate its fitness.

        Parameters
        ----------
        new_state : np.ndarray
            New state vector.
        """
        if len(new_state) != self.length:
            raise ValueError(f"new_state length {len(new_state)} must match problem length {self.length}")

        self.state = new_state
        self.fitness = self.eval_fitness(self.state)

    def can_stop(self) -> bool:
        """Determine if the optimization process can stop.

        Returns
        -------
        bool
            Always returns False for this base class.
        """
        return False



"""Classes for defining fitness functions (i.e., optimization problems) for optimization algorithms."""

from .continuous_peaks import ContinuousPeaks
from .flip_flop import FlipFlop
from .four_peaks import FourPeaks
from .knapsack import Knapsack
from .max_k_color import MaxKColor
from .one_max import OneMax
from .queens import Queens
from .six_peaks import SixPeaks
from .travelling_sales import TravellingSales

from .custom_fitness import CustomFitness



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any, Optional, Callable

import numpy as np

from mlrose_ky.neural._nn_base import _NNBase
from mlrose_ky.neural.activation import tanh


class NNClassifier(_NNBase):
    """
    Neural Network Classifier class based on _NNBase.

    Parameters
    ----------
    runner : Any
        Object that controls the execution of training experiments.
    algorithm : str, optional
        The optimization algorithm to use for training (e.g., "random_hill_climb", "simulated_annealing", etc.).
    activation : Callable, optional, default=mlrose_ky.tanh
        Activation function to use in the neural network layers.
    hidden_layer_sizes : list[int], optional
        List defining the number of hidden layers and nodes in each layer.
    max_iters : int, default=100
        Maximum number of iterations for the optimization algorithm.
    max_attempts : int, default=10
        Maximum number of attempts to find a better state at each step.
    learning_rate_init : float, default=0.1
        Initial learning rate for optimization.
    bias : bool, default=True
        Whether to include bias nodes in the network.
    early_stopping : bool, default=False
        Whether to stop training early if no improvement is detected.
    clip_max : float, default=1e10
        Maximum value for clipping weights during optimization.
    seed : int or None, default=None
        Random seed for reproducibility.
    kwargs : dict, optional
        Additional arguments passed to the training functions.
    """

    def __init__(
        self,
        runner: Any,
        algorithm: Any = None,
        activation: Optional[Callable] = tanh,
        hidden_layer_sizes: list[int] = None,
        max_iters: int = 100,
        max_attempts: int = 10,
        learning_rate_init: float = 0.1,
        bias: bool = True,
        early_stopping: bool = False,
        clip_max: float = 1e10,
        seed: int = None,
        **kwargs: Any,
    ):
        super().__init__()

        self.runner = runner
        self.grid_search_parameters = runner.grid_search_parameters

        # Neural network properties
        self.hidden_layer_sizes: list[int] = hidden_layer_sizes if hidden_layer_sizes is not None else []

        self.activation: Optional[Callable] = activation
        self.learning_rate_init: float = learning_rate_init
        self.max_iters: int = max_iters
        self.max_attempts: int = max_attempts

        self.bias: bool = bias
        self.early_stopping: bool = early_stopping
        self.clip_max: float = clip_max
        self.algorithm: Any | None = algorithm

        # Result properties
        self.fitness_fn: Any = None
        self.problem: Any = None
        self.fitted_weights: np.ndarray | None = None
        self.output_activation: Optional[Callable] = None
        self.predicted_probabilities: np.ndarray | None = None
        self.node_list: list[int] | None = None
        self.node_count: int | None = None
        self.loss: float | None = None
        self.fit_started_: bool = False
        self.seed: int | None = seed

        # Extra parameters
        self.kwargs: dict[str, Any] = kwargs
        for k, v in kwargs.items():
            if hasattr(self, k):
                # Ignore kwargs that are already defined as attributes
                continue
            self.__setattr__(k, v)

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """
        Get the parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, returns the parameters for this estimator and its sub-objects.

        Returns
        -------
        dict
            Parameter names mapped to their values.
        """
        out = super().get_params(deep)

        # Exclude any attributes that end with an underscore
        out = {k: v for (k, v) in out.items() if not k.endswith("_")}
        ap = {k: self.__dict__.get(k, None) for k in self.grid_search_parameters if k not in out}
        out.update(ap)

        return out

    def _get_nodes(self, x_train: np.ndarray, y_train: np.ndarray) -> list[int]:
        """
        Get the number of nodes in each layer of the network.

        Parameters
        ----------
        x_train : np.ndarray
            The training feature set.
        y_train : np.ndarray
            The training label set.

        Returns
        -------
        list of int
            Number of nodes in each layer.
        """
        return _NNBase._build_node_list(X=x_train, y=y_train, hidden_nodes=self.hidden_layer_sizes, bias=self.bias)

    def fit(self, x_train: np.ndarray, y_train: np.ndarray = None, init_weights: np.ndarray = None) -> "NNClassifier":
        """
        Fit the neural network classifier to the training data.

        Parameters
        ----------
        x_train : np.ndarray
            The training feature set.
        y_train : np.ndarray, optional
            The training label set.
        init_weights : np.ndarray, optional
            Initial weights for the network. If None, random weights are used.

        Returns
        -------
        NNClassifier
            The fitted classifier.
        """
        x_train, y_train = self._format_x_y_data(x_train, y_train)
        self.node_list = self._get_nodes(x_train, y_train)
        self.node_count = _NNBase._calculate_state_size(self.node_list)

        fitness, problem = _NNBase._build_problem_and_fitness_function(
            X=x_train,
            y=y_train,
            node_list=self.node_list,
            activation=self.activation,
            learning_rate=self.learning_rate_init,
            clip_max=self.clip_max,
            bias=self.bias,
        )
        self.fitness_fn = fitness
        self.problem = problem

        # If algorithm is None, skip the training process
        if self.algorithm is None:
            self.fitted_weights = None
            self.loss = None
            self.output_activation = None
            return self

        # Check for early abort or replay mode.
        if self.runner.has_aborted() or self.runner.replay_mode():
            self.fitted_weights = np.array([np.NaN] * self.node_count)
            self.loss = np.NaN
            self.output_activation = self.fitness_fn.get_output_activation()
            return self

        # Handle grid search or regular training
        params = {k: self.__getattribute__(k) for k in self.kwargs}
        if init_weights is None:
            np.random.seed(self.seed)
            init_weights = np.random.uniform(-1, 1, self.node_count)

        params["init_state"] = init_weights
        total_args = {
            "algorithm": self.algorithm,
            "activation": self.activation,
            "bias": self.bias,
            "early_stopping": self.early_stopping,
            "clip_max": self.clip_max,
            "hidden_layer_sizes": self.hidden_layer_sizes,
            "learning_rate_init": self.learning_rate_init,
        }
        max_attempts = self.max_attempts if self.early_stopping else self.max_iters
        self.fit_started_ = True

        fitted_weights, loss, _ = self.runner.run_one_experiment_(
            algorithm=self.algorithm, problem=problem, max_iters=self.max_iters, max_attempts=max_attempts, total_args=total_args, **params
        )

        # Save fitted weights
        self.fitted_weights = problem.get_state()
        self.loss = loss
        self.output_activation = self.fitness_fn.get_output_activation()

        return self

    def predict(self, x_test: np.ndarray) -> np.ndarray:
        """
        Predict labels for the test feature set.

        Parameters
        ----------
        x_test : np.ndarray
            The test feature set.

        Returns
        -------
        np.ndarray
            Predicted labels.
        """
        if np.shape(x_test)[1] != (self.node_list[0] - self.bias):
            raise ValueError(f"The number of columns in X must equal {self.node_list[0] - self.bias}, got {np.shape(x_test)[1]}.")

        y_pred, pp = self._predict(
            X=x_test,
            fitted_weights=self.fitted_weights,
            node_list=self.node_list,
            input_activation=self.activation,
            output_activation=self.output_activation,
            bias=self.bias,
        )

        self.predicted_probabilities = pp

        return y_pred

    def predict_proba(self, x_test: np.ndarray) -> np.ndarray:
        """
        Predict class probabilities for the test feature set.

        Parameters
        ----------
        x_test : np.ndarray
            The test feature set.

        Returns
        -------
        np.ndarray
            Predicted probabilities.
        """
        self.predict(x_test)
        return self.predicted_probabilities



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np

from mlrose_ky.algorithms.crossovers import UniformCrossOver
from mlrose_ky.algorithms.mutators import ChangeOneMutator
from mlrose_ky.fitness.queens import Queens
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class QueensOpt(DiscreteOpt):
    """Class for defining the N-Queens optimization problem.

    Parameters
    ----------
    length : int, default=None
        Number of queens (and the size of the board). If not provided, it is inferred from the fitness function.

    fitness_fn : Queens, default=None
        Object to implement the fitness function for optimization. If not provided, a new `Queens` function will be created.

    maximize : bool, default=False
        Whether to maximize the fitness function. Set :code:`False` for minimization problems.

    crossover : UniformCrossOver, default=None
        Crossover operation used for reproduction. If None, defaults to `UniformCrossOver`.

    mutator : ChangeOneMutator, default=None
        Mutation operation used for reproduction. If None, defaults to `ChangeOneMutator`.

    Attributes
    ----------
    length : int
        Number of queens and size of the board.

    fitness_fn : Queens
        Fitness function for the optimization problem.

    max_val : int
        Maximum value for each element in the state vector (equals to the length).

    stop_fitness : int
        The fitness value at which the optimization process can stop.
    """

    def __init__(
        self,
        length: int = None,
        fitness_fn: Any = None,
        maximize: bool = False,
        crossover: "UniformCrossOver" = None,
        mutator: "ChangeOneMutator" = None,
    ):
        # Ensure that either fitness_fn or length is provided
        if fitness_fn is None and length is None:
            raise ValueError("Either fitness_fn or length must be specified.")

        # Infer length from fitness_fn if not provided
        if length is None:
            length = len(fitness_fn.weights)

        self.length: int = length

        # If fitness_fn is not provided, create a new Queens fitness function
        if fitness_fn is None:
            fitness_fn = Queens(maximize=maximize)

        # Set the stopping fitness value based on whether we're maximizing or minimizing
        self.stop_fitness: int = Queens.get_max_size(length) if maximize else 0

        # Set max_val to length, as it represents the board size
        self.max_val: int = length

        # Use default crossover and mutator if none are provided
        crossover = UniformCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator
        super().__init__(length, fitness_fn, maximize, length, crossover, mutator)

        # Initialize the state with a random, shuffled assignment of queens
        state = np.random.randint(self.length, size=self.length)
        np.random.shuffle(state)
        self.set_state(state)

    def can_stop(self) -> bool:
        """Determine if the optimization process can stop.

        Returns
        -------
        bool
            True if the fitness value matches the stopping criterion, otherwise False.
        """
        return int(self.get_fitness()) == self.stop_fitness



"""mlrose-ky initialization file."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

# noinspection PyUnresolvedReferences
from .algorithms.ga import genetic_alg

# noinspection PyUnresolvedReferences
from .algorithms.sa import simulated_annealing

# noinspection PyUnresolvedReferences
from .algorithms.hc import hill_climb

# noinspection PyUnresolvedReferences
from .algorithms.rhc import random_hill_climb

# noinspection PyUnresolvedReferences
from .algorithms.gd import gradient_descent

# noinspection PyUnresolvedReferences
from .algorithms.mimic import mimic

# noinspection PyUnresolvedReferences
from .algorithms.decay import GeomDecay, ArithDecay, ExpDecay, CustomSchedule

# noinspection PyUnresolvedReferences
from .algorithms.crossovers import OnePointCrossOver, UniformCrossOver, TSPCrossOver

# noinspection PyUnresolvedReferences
from .algorithms.mutators import ChangeOneMutator, DiscreteMutator, SwapMutator, ShiftOneMutator

# noinspection PyUnresolvedReferences
from .fitness import OneMax, FlipFlop, FourPeaks, SixPeaks, ContinuousPeaks, Knapsack, TravellingSales, Queens, MaxKColor, CustomFitness

# noinspection PyUnresolvedReferences,PyProtectedMember
from .neural import NeuralNetwork, LinearRegression, LogisticRegression, NNClassifier, _nn_core

# noinspection PyUnresolvedReferences
from .neural.activation import identity, relu, leaky_relu, sigmoid, softmax, tanh

# noinspection PyUnresolvedReferences
from .neural.fitness import NetworkWeights

# noinspection PyUnresolvedReferences
from .neural.utils.weights import flatten_weights, unflatten_weights

# noinspection PyUnresolvedReferences
from .gridsearch import GridSearchMixin

# noinspection PyUnresolvedReferences
from .opt_probs import DiscreteOpt, ContinuousOpt, KnapsackOpt, TSPOpt, QueensOpt, FlipFlopOpt, MaxKColorOpt

# noinspection PyUnresolvedReferences
from .runners import GARunner, MIMICRunner, RHCRunner, SARunner, NNGSRunner, SKMLPRunner, build_data_filename

# noinspection PyUnresolvedReferences
from .generators import MaxKColorGenerator, QueensGenerator, FlipFlopGenerator, TSPGenerator, KnapsackGenerator, ContinuousPeaksGenerator

# noinspection PyUnresolvedReferences
from .samples import SyntheticData, plot_synthetic_dataset



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

from mlrose_ky.algorithms.crossovers import UniformCrossOver
from mlrose_ky.algorithms.mutators import ChangeOneMutator
from mlrose_ky.fitness.knapsack import Knapsack
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class KnapsackOpt(DiscreteOpt):
    """Class for defining Knapsack optimization problems.

    Parameters
    ----------
    length : int, default=None
        Number of elements in the state vector. If not specified, it is inferred from the weights, values, or fitness function.

    fitness_fn : Knapsack, default=None
        Object to implement the fitness function for optimization. If not specified, a new `Knapsack` fitness function will be created.

    maximize : bool, default=True
        Whether to maximize the fitness function. Set :code:`False` for minimization problems.

    max_val : int, default=2
        Maximum number of items that can be selected for each element in the state vector.

    weights : list[float] | None, default=None
        List of weights for each item. Required if `fitness_fn` is not provided.

    values : list[float] | None, default=None
        List of values for each item. Required if `fitness_fn` is not provided.

    max_weight_pct : float, default=0.35
        Maximum allowable weight as a percentage of the total weight.

    crossover : UniformCrossOver, default=None
        Crossover operation used for reproduction. If None, defaults to `UniformCrossOver`.

    mutator : ChangeOneMutator, default=None
        Mutation operation used for reproduction. If None, defaults to `ChangeOneMutator`.

    multiply_by_max_item_count : bool, default=False
        Whether to multiply the values by the maximum item count allowed.

    Attributes
    ----------
    length : int
        Number of elements in the state vector.

    fitness_fn : Knapsack
        Fitness function for the optimization problem.

    max_val : int
        Maximum number of items that can be selected for each element in the state vector.
    """

    def __init__(
        self,
        length: int = None,
        fitness_fn: Any = None,
        maximize: bool = True,
        max_val: int = 2,
        weights: list[float] = None,
        values: list[float] = None,
        max_weight_pct: float = 0.35,
        crossover: "UniformCrossOver" = None,
        mutator: "ChangeOneMutator" = None,
        multiply_by_max_item_count: bool = False,
    ):
        if fitness_fn is None and (weights is None and values is None):
            raise ValueError("Either fitness_fn or both weights and values must be specified.")

        if length is None:
            if weights is not None:
                length = len(weights)
            elif values is not None:
                length = len(values)
            elif fitness_fn is not None:
                length = len(fitness_fn.weights)

        self.length: int = length
        self.max_val: int = max_val

        if fitness_fn is None:
            fitness_fn = Knapsack(
                weights=weights,
                values=values,
                max_weight_pct=max_weight_pct,
                max_item_count=max_val,
                multiply_by_max_item_count=multiply_by_max_item_count,
            )

        crossover = UniformCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator

        super().__init__(length, fitness_fn, maximize, max_val, crossover, mutator)



"""
Provides decorators to assign and retrieve short names for functions and methods.

This file contains decorators that allow for the easy assignment of a 'short name'
to functions and methods, which can be particularly useful for assigning an algorithm's
common or shortened name, such as 'GA' for 'Genetic Algorithms'.

It also includes a function to retrieve the assigned short name of a function.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Callable, Any


def short_name(expr: str) -> Callable:
    """
    Decorator to assign a short name to a function.

    Parameters
    ----------
    expr : str
        The short name to be assigned to the function.

    Returns
    -------
    Callable
        A decorator that assigns the provided short name to a function and returns the function.
    """

    def short_name_func_applicator(func: Callable) -> Callable:
        """Assign a short name to the given function."""
        func.__short_name__ = expr
        return func

    return short_name_func_applicator


def get_short_name(v: Any) -> str:
    """
    Retrieve the short name of a variable, or its default name if a short name isn't assigned.

    Parameters
    ----------
    v : Any
        The variable from which the short name is retrieved.

    Returns
    -------
    str
        The short name of the variable, if assigned; otherwise, returns the full variable name or the variable itself as a fallback.
    """
    if hasattr(v, "__short_name__"):
        return v.__short_name__
    elif hasattr(v, "__name__"):
        return v.__name__
    return v



"""Classes for defining crossover strategies for Genetic Algorithms (GA)."""

# Author: Genevieve Hayes
# License: BSD 3-clause

from .one_point_crossover import OnePointCrossOver
from .tsp_crossover import TSPCrossOver
from .uniform_crossover import UniformCrossOver



"""Classes for defining optimization problem objects."""

from typing import Any

import numpy as np

from mlrose_ky.opt_probs._opt_prob import _OptProb


# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause


class ContinuousOpt(_OptProb):
    """Class for defining continuous-state optimization problems.

    Parameters
    ----------
    length : int
        Number of elements in the state vector.

    fitness_fn : Any
        Fitness function for optimization.

    maximize : bool, default=True
        Whether to maximize the fitness function.
        Set :code:`False` for minimization problem.

    min_val : float, default=0
        Minimum value that each element of the state vector can take.

    max_val : float, default=1
        Maximum value that each element of the state vector can take.

    step : float, default=0.
        Step size used in determining neighbors of the current state.

    Attributes
    ----------
    min_val : float
        Minimum value for the state vector elements.
    max_val : float
        Maximum value for the state vector elements.
    step : float
        Step size used in neighbor determination.
    prob_type : str
        Problem type; always 'continuous' for this class.
    """

    def __init__(self, length: int, fitness_fn: Any, maximize: bool = True, min_val: float = 0.0, max_val: float = 1.0, step: float = 0.1):
        super().__init__(length, fitness_fn, maximize=maximize)

        if self.fitness_fn.get_prob_type() not in {"continuous", "either"}:
            raise ValueError(
                "fitness_fn must have problem type 'continuous' or 'either'. "
                "Define problem as DiscreteOpt or use an appropriate fitness function."
            )

        if max_val <= min_val:
            raise ValueError("max_val must be greater than min_val.")
        if step <= 0:
            raise ValueError("step size must be positive.")
        if (max_val - min_val) < step:
            raise ValueError(f"step size must be less than (max_val - min_val).")

        self.prob_type: str = "continuous"
        self.min_val: float = min_val
        self.max_val: float = max_val
        self.step: float = step

    def calculate_updates(self) -> list:
        """Calculate gradient descent updates.

        Returns
        -------
        list
            List of back-propagation weight updates.
        """
        return self.fitness_fn.calculate_updates()

    def find_neighbors(self):
        """Find all neighbors of the current state."""
        # Pre-allocate a NumPy array for neighbors (maximum of 2 * length neighbors)
        neighbors_matrix = np.zeros((2 * self.length, self.length))

        neighbor_count = 0  # Track how many valid neighbors we find

        for i in range(self.length):
            base_neighbor = np.copy(self.state)

            for step_dir in [-self.step, self.step]:
                neighbor = base_neighbor.copy()
                neighbor[i] += step_dir

                # Clip the value to ensure it remains within bounds
                neighbor[i] = np.clip(neighbor[i], self.min_val, self.max_val)

                # Only add if different from the current state
                if not np.array_equal(neighbor, self.state):
                    neighbors_matrix[neighbor_count] = neighbor
                    neighbor_count += 1

        # Store only the valid neighbors (slice the array up to the count of valid ones)
        self.neighbors = neighbors_matrix[:neighbor_count]

    def get_prob_type(self) -> str:
        """Return the problem type.

        Returns
        -------
        str
            Problem type.
        """
        return self.prob_type

    def random(self) -> np.ndarray:
        """Generate and return a random state vector.

        Returns
        -------
        np.ndarray
            Randomly generated state vector.
        """
        return np.random.uniform(self.min_val, self.max_val, self.length)

    def random_neighbor(self) -> np.ndarray:
        """Return a random neighbor of the current state vector.

        Returns
        -------
        np.ndarray
            State vector of the random neighbor.
        """
        while True:
            neighbor = np.copy(self.state)
            i = np.random.randint(0, self.length)
            neighbor[i] += self.step * np.random.choice([-1, 1])

            if neighbor[i] > self.max_val:
                neighbor[i] = self.max_val
            elif neighbor[i] < self.min_val:
                neighbor[i] = self.min_val

            if not np.array_equal(neighbor, self.state):
                break

        return neighbor

    def random_pop(self, pop_size: int):
        """Create a population of random state vectors.

        Parameters
        ----------
        pop_size : int
            Size of the population to be created.

        Raises
        ------
        ValueError
            If pop_size is not a positive integer.
        """
        if pop_size <= 0 or not isinstance(pop_size, int):
            raise ValueError("pop_size must be a positive integer.")

        population = []
        pop_fitness = []

        for _ in range(pop_size):
            state = self.random()
            fitness = self.eval_fitness(state)
            population.append(state)
            pop_fitness.append(fitness)

        self.population = np.array(population)
        self.pop_fitness = np.array(pop_fitness)

    def reproduce(self, parent_1: np.ndarray, parent_2: np.ndarray, mutation_prob: float = 0.1) -> np.ndarray:
        """Create a child state vector from two parent state vectors.

        Parameters
        ----------
        parent_1 : np.ndarray
            State vector for parent 1.

        parent_2 : np.ndarray
            State vector for parent 2.

        mutation_prob : float, default=0.1
            Probability of a mutation at each state vector element during reproduction.

        Returns
        -------
        np.ndarray
            Child state vector produced from parents 1 and 2.

        Raises
        ------
        ValueError
            If the lengths of the parents do not match the problem length,
            or if mutation_prob is not between 0 and 1.
        """
        if len(parent_1) != self.length or len(parent_2) != self.length:
            raise ValueError("Lengths of parents must match problem length.")

        if not (0 <= mutation_prob <= 1):
            raise ValueError("mutation_prob must be between 0 and 1.")

        if self.length > 1:
            _n = np.random.randint(self.length - 1)
            child = np.zeros(self.length)
            child[0 : _n + 1] = parent_1[0 : _n + 1]
            child[_n + 1 :] = parent_2[_n + 1 :]
        else:
            child = np.copy(parent_1 if np.random.randint(2) == 0 else parent_2)

        # Mutate child
        rand = np.random.uniform(size=self.length)
        mutate = np.where(rand < mutation_prob)[0]

        for i in mutate:
            child[i] = np.random.uniform(self.min_val, self.max_val)

        return child

    def reset(self):
        """Set the current state vector to a random value and reset its fitness."""
        self.state = self.random()
        self.fitness_evaluations = 0
        self.fitness = self.eval_fitness(self.state)

    def update_state(self, updates: np.ndarray) -> np.ndarray:
        """Update the current state given a vector of updates.

        Parameters
        ----------
        updates : np.ndarray
            Array of updates.

        Returns
        -------
        np.ndarray
            Current state adjusted for updates.

        Raises
        ------
        ValueError
            If the length of updates does not match the problem length.
        """
        if len(updates) != self.length:
            raise ValueError("Length of updates must match problem length.")

        updated_state = self.state + updates
        updated_state[updated_state > self.max_val] = self.max_val
        updated_state[updated_state < self.min_val] = self.min_val

        return updated_state



"""Unit tests for generators/"""

# Author: Kyle Nakamura
# License: BSD 3-clause

from unittest.mock import patch
import matplotlib.pyplot as plt
import pytest
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier

from tests.globals import SEED
from mlrose_ky.samples import SyntheticData, plot_synthetic_dataset


# noinspection PyMissingOrEmptyDocstring
@pytest.fixture
def generator():
    return SyntheticData(seed=SEED)


class TestSyntheticDataGenerator:
    def test_get_synthetic_features_and_classes_default(self, generator):
        """Test default synthetic features and classes."""
        features, classes = generator.get_synthetic_features_and_classes()

        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]

    def test_get_synthetic_features_and_classes_with_redundant(self, generator):
        """Test synthetic features and classes with redundant column."""
        features, classes = generator.get_synthetic_features_and_classes(with_redundant_column=True)

        assert features == ["(1) A", "(2) B", "(3) R"]
        assert classes == ["RED", "BLUE"]

    def test_get_synthetic_data(self, generator):
        """Test getting synthetic data."""
        data, features, classes, output_directory = generator.get_synthetic_data()

        assert data.shape == (400, 3)
        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is None

    def test_setup_synthetic_data_test_train(self, generator):
        """Test setting up synthetic data for train and test."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)

        assert x_train.shape == (280, 2)
        assert x_test.shape == (120, 2)
        assert y_train.shape == (280,)
        assert y_test.shape == (120,)

    @patch("matplotlib.pyplot.show")  # Mocking plt.show
    def test_plot_synthetic_dataset(self, mock_show, generator):
        """Test plotting synthetic dataset."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        classifier = LogisticRegression().fit(x_train, y_train)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test, classifier=classifier)
        plt.close()

        assert mock_show.called

    @patch("matplotlib.pyplot.show")
    def test_plot_synthetic_dataset_with_predict_proba(self, mock_show, generator):
        """Test plotting synthetic dataset with a classifier that uses predict_proba."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        classifier = RandomForestClassifier(random_state=generator.seed).fit(x_train, y_train)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test, classifier=classifier)
        plt.close()

        assert mock_show.called

    def test_get_synthetic_data_with_noise(self, generator):
        """Test getting synthetic data with noise."""
        data, features, classes, output_directory = generator.get_synthetic_data(add_noise=0.1)

        assert data.shape[0] > 400  # Noise adds more data points
        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is None

    def test_get_synthetic_data_with_redundant_column(self, generator):
        """Test getting synthetic data with a redundant column."""
        data, features, classes, output_directory = generator.get_synthetic_data(add_redundant_column=True)

        assert data.shape == (400, 4)  # Additional column for redundancy
        assert features == ["(1) A", "(2) B", "(3) R"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is None

    @patch("mlrose_ky.samples.synthetic_data.makedirs")
    def test_get_synthetic_data_with_root_directory(self, mock_makedirs, generator):
        """Test getting synthetic data with a root directory."""
        generator.root_directory = "/tmp/synthetic_data"
        data, features, classes, output_directory = generator.get_synthetic_data()

        assert data.shape == (400, 3)
        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is not None
        mock_makedirs.assert_called_once_with(output_directory)

    @patch("mlrose_ky.samples.synthetic_data.makedirs")
    def test_get_synthetic_data_with_root_directory_oserror(self, mock_makedirs, generator):
        """Test getting synthetic data with a root directory where makedirs raises OSError."""
        # Arrange: Set the side effect of makedirs to raise OSError
        mock_makedirs.side_effect = OSError("Test OSError")

        # Act: Set the root directory and call get_synthetic_data
        generator.root_directory = "/tmp/synthetic_data"
        data, features, classes, output_directory = generator.get_synthetic_data()

        # Assert: Check that data is returned correctly even when OSError is raised
        assert data.shape == (400, 3)
        assert features == ["(1) A", "(2) B"]
        assert classes == ["RED", "BLUE"]
        assert output_directory is not None
        mock_makedirs.assert_called_once_with(output_directory)

    @patch("matplotlib.pyplot.show")
    def test_plot_synthetic_dataset_without_classifier(self, mock_show, generator):
        """Test plotting synthetic dataset without a classifier."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test)
        plt.close()

        assert mock_show.called

    @patch("matplotlib.pyplot.show")
    def test_plot_synthetic_dataset_with_redundant_column(self, mock_show, generator):
        """Test plotting synthetic dataset with a redundant column (three features)."""
        data, _, _, _ = generator.get_synthetic_data(add_redundant_column=True)
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        classifier = LogisticRegression().fit(x_train, y_train)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test, classifier=classifier)
        plt.close()

        assert mock_show.called

    @patch("matplotlib.pyplot.show")
    def test_plot_synthetic_dataset_with_transparent_bg(self, mock_show, generator):
        """Test plotting synthetic dataset with transparent background and custom background color."""
        data, _, _, _ = generator.get_synthetic_data()
        x, y, x_train, x_test, y_train, y_test = generator.setup_synthetic_data_test_train(data)
        classifier = LogisticRegression().fit(x_train, y_train)
        plt.figure()
        plot_synthetic_dataset(x_train, x_test, y_train, y_test, classifier=classifier, transparent_bg=True, bg_color="black")
        plt.close()

        assert mock_show.called



"""Neural network 'Leaky ReLu' activation function."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# Contributor : Ankit Grover
# License: BSD 3-clause

import warnings

import numpy as np

from mlrose_ky.decorators import short_name

warnings.filterwarnings("ignore")


@short_name("leaky_relu")
def leaky_relu(x: np.ndarray, alpha: float = 0.3, deriv: bool = False) -> np.ndarray:
    """Leaky ReLU activation function.

    Parameters
    ----------
    x: np.ndarray
        Array containing input data.
    alpha: int, optional, default=0.3
        Value to be set for applying small negative gradient.
    deriv: bool, option, default=False
        Whether to return the function (when False) or its derivative (when True).

    Returns
    -------
    fx: np.ndarray
        Value of activation function at x.
    """
    fx = np.copy(x)
    fx = np.where(fx < 0, fx * alpha, fx)

    if deriv:
        fx[np.where(fx > 0)] = 1
        fx[np.where(fx < 0)] = alpha

    return fx



"""Crossover implementations for Genetic Algorithms (GA).

This module defines a base class for crossover operations used in genetic algorithms,
detailing how two parent solutions can be combined to create offspring.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from abc import ABC, abstractmethod
from typing import Any, Sequence

import numpy as np


class _CrossOverBase(ABC):
    """
    Base class for crossover operations in a genetic algorithm.

    Provides a structured way to define crossover behavior in genetic algorithms.
    It should be subclassed to implement specific crossover strategies.

    Parameters
    ----------
    opt_prob : Any
        An instance of the optimization problem related to the genetic algorithm.
        This problem instance should provide necessary properties like 'length'
        that might be needed for the crossover operation.

    Attributes
    ----------
    _opt_prob : Any
        The optimization problem instance.
    _length : int
        Length of the chromosome, typically derived from the optimization problem's
        'length' property.
    """

    def __init__(self, opt_prob: Any):
        """
        Initialize the _CrossOverBase with the given optimization problem.

        Parameters
        ----------
        opt_prob : Any
            An instance of the optimization problem related to the GA.
        """
        super().__init__()
        self._opt_prob: Any = opt_prob
        self._length: int = opt_prob.length

    @abstractmethod
    def mate(self, p1: Sequence[int | float], p2: Sequence[int | float]) -> np.ndarray:
        """
        Perform the crossover (mating) between two parents to produce offspring.

        This method must be implemented by subclasses to define specific crossover
        behavior based on the genetics of the parents.
        """
        raise NotImplementedError("Subclasses must implement this method")



"""Mutation implementations for Genetic Algorithms (GA).

This module defines a base class for mutation operations used in genetic algorithms,
detailing how individual solutions (chromosomes) can be altered to introduce genetic diversity.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from abc import ABC, abstractmethod
from typing import Any

import numpy as np


class _MutatorBase(ABC):
    """
    Base class for mutation operations in a genetic algorithm.

    This class provides a structured way to define mutation operations that
    can be applied to genetic algorithm populations.

    Attributes
    ----------
    _opt_prob : Any
        The optimization problem instance associated with the mutation operations.
    _length : int
        The length of the chromosome in the genetic algorithm, derived from the optimization problem.

    Parameters
    ----------
    opt_prob : Any
        An instance of an optimization problem that the mutator will operate on.
    """

    def __init__(self, opt_prob: Any):
        super().__init__()

        self._opt_prob: Any = opt_prob
        self._length: int = opt_prob.length

    @abstractmethod
    def mutate(self, child: np.ndarray, mutation_probability: float) -> np.ndarray:
        """
        Apply mutation operation to a given child chromosome based on a mutation probability.

        This method must be implemented by subclasses to define specific mutation strategies.
        """
        raise NotImplementedError("Subclasses must implement this method")



"""Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes
# License: BSD 3-clause


from typing import Callable

import numpy as np
import sklearn.metrics as skm

from mlrose_ky.neural import activation as act
from mlrose_ky.neural.utils import unflatten_weights


class NetworkWeights:
    """
    Fitness function for neural network weights optimization problem.

    Parameters
    ----------
    X : np.ndarray
        Numpy array containing feature dataset with each row representing a
        single observation.
    y : np.ndarray
        Numpy array containing true values of data labels.
        Length must be the same as the length of X.
    node_list : list[int]
        Number of nodes in each layer, including the input and output layers.
    activation : Callabe
        Activation function for each of the hidden layers with the signature
        `activation(x, deriv)`, where setting deriv is a boolean that
        determines whether to return the activation function or its derivative.
    bias : bool, default=True
        Whether a bias term is included in the network.
    is_classifier : bool, default=True
        Whether the network is for classification or regression. Set True for
        classification and False for regression.
    learning_rate : float, default=0.1
        The learning rate for gradient descent updates.
    """

    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        node_list: list[int],
        activation: Callable,
        bias: bool = True,
        is_classifier: bool = True,
        learning_rate: float = 0.1,
    ):
        if not callable(activation):
            raise TypeError("Activation function must be callable.")
        try:
            activation(np.array([0.1]), deriv=False)
        except TypeError:
            raise TypeError("Activation function must accept two arguments: 'x' and 'deriv'.")

        if X.size == 0 or y.size == 0:
            raise ValueError("X and y cannot be empty.")

        y = np.array(y)

        if len(np.shape(y)) == 1:
            y = np.reshape(y, [len(y), 1])

        if np.shape(X)[0] != np.shape(y)[0]:
            raise ValueError(f"The length of X ({np.shape(X)[0]}) and y ({np.shape(y)[0]}) must be equal.")

        if len(node_list) < 2:
            raise ValueError("node_list must contain at least 2 elements.")

        if np.shape(X)[1] != (node_list[0] - bias):
            raise ValueError(f"The number of columns in X must equal {node_list[0] - bias}.")

        if np.shape(y)[1] != node_list[-1]:
            raise ValueError(f"The number of columns in y must equal {node_list[-1]}.")

        if not isinstance(bias, bool):
            raise ValueError("bias must be True or False.")

        if not isinstance(is_classifier, bool):
            raise ValueError("is_classifier must be True or False.")

        if learning_rate <= 0:
            raise ValueError("learning_rate must be greater than 0.")

        self.X = X
        self.y_true = y
        self.node_list = node_list
        self.activation = activation
        self.bias = bias
        self.is_classifier = is_classifier
        self.learning_rate = learning_rate

        if self.is_classifier:
            self.loss = skm.log_loss
            self.output_activation = act.sigmoid if np.shape(self.y_true)[1] == 1 else act.softmax
        else:
            self.loss = skm.mean_squared_error
            self.output_activation = act.identity

        self.inputs_list = []
        self.y_pred = y
        self.weights = []
        self.prob_type = "continuous"

        self.nodes = sum(node_list[i] * node_list[i + 1] for i in range(len(node_list) - 1))

    def evaluate(self, state: np.ndarray) -> float:
        """
        Evaluate the fitness of a state.

        Parameters
        ----------
        state : np.ndarray
            State array for evaluation.

        Returns
        -------
        fitness : float
            Value of fitness function.
        """
        if len(state) != self.nodes:
            raise ValueError(f"state must have length {self.nodes}, got {len(state)}.")

        self.inputs_list = []
        self.weights = list(unflatten_weights(state, self.node_list))

        if self.bias:
            ones = np.ones([np.shape(self.X)[0], 1])
            inputs = np.hstack((self.X, ones))
        else:
            inputs = self.X

        for i in range(len(self.weights)):
            outputs = np.dot(inputs, self.weights[i])
            self.inputs_list.append(inputs)

            inputs = self.activation(outputs) if i < len(self.weights) - 1 else self.output_activation(outputs)

        self.y_pred = inputs
        return self.loss(self.y_true, self.y_pred)

    def get_output_activation(self) -> Callable:
        """
        Return the activation function for the output layer.

        Returns
        -------
        Callable
            Activation function for the output layer.
        """
        return self.output_activation

    def get_prob_type(self) -> str:
        """
        Return the problem type.

        Returns
        -------
        str
            Problem type as 'discrete', 'continuous', 'tsp', or 'either'.
        """
        return self.prob_type

    def calculate_updates(self) -> list[np.ndarray]:
        """
        Calculate gradient descent updates.

        Returns
        -------
        list of np.ndarray
            List of back propagation weight updates.
        """
        delta_list: list = []
        updates_list: list = []

        for i in range(len(self.inputs_list) - 1, -1, -1):
            if i == len(self.inputs_list) - 1:
                delta = self.y_pred - self.y_true
            else:
                dot = np.dot(delta_list[-1], np.transpose(self.weights[i + 1]))
                activation = self.activation(self.inputs_list[i + 1], deriv=True)
                delta = dot * activation

            delta_list.append(delta)
            updates = -self.learning_rate * np.dot(np.transpose(self.inputs_list[i]), delta)
            updates_list.append(updates)

        return updates_list[::-1]



"""
Class for running optimization experiments using Genetic Algorithms (GA), including grid search functionality.

Example usage:
    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    ga = GARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(12),
                  max_attempts=1000,
                  population_sizes=[150, 200, 300],
                  mutation_rates=[0.4, 0.5, 0.6])

    df_run_stats, df_run_curves = ga.run()
"""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np
import pandas as pd

from mlrose_ky.algorithms import genetic_alg
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase


@short_name("ga")
class GARunner(_RunnerBase):
    """
    A runner for performing optimization experiments using Genetic Algorithms (GA).

    This class extends _RunnerBase and provides functionality for running experiments
    with Genetic Algorithms, including grid search over hyperparameters such as
    population size and mutation rate.

    Attributes
    ----------
    population_sizes : list[int]
        List of population sizes to test in the grid search.
    mutation_rates : list[float]
        List of mutation rates to test in the grid search.
    hamming_factors : Optional[list[float]]
        Optional list of Hamming factors for the algorithm.
    hamming_factor_decays : Optional[list[float]]
        Optional list of Hamming factor decays.
    """

    def __init__(
        self,
        problem: Any,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        population_sizes: list[int],
        mutation_rates: list[float],
        hamming_factors: list[float] = None,
        hamming_factor_decays: list[float] = None,
        max_attempts: int = 500,
        generate_curves: bool = True,
        output_directory: str = None,
        **kwargs: Any,
    ):
        """
        Initialize the GARunner class with problem data and various experiment parameters.

        Parameters
        ----------
        problem : Any
            The optimization problem to be solved.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed for reproducibility.
        iteration_list : np.ndarray | list of int
            List of iterations for the experiment.
        population_sizes : list of int
            List of population sizes to test in the grid search.
        mutation_rates : list of float
            List of mutation rates to test in the grid search.
        hamming_factors : list of float, optional
            Optional list of Hamming factors for the algorithm.
        hamming_factor_decays : list of float, optional
            Optional list of Hamming factor decays.
        max_attempts : int, optional
            Maximum number of attempts without improvement before stopping.
        generate_curves : bool, optional
            Whether to generate learning curves.
        output_directory : str, optional
            Directory to save experiment result, default=None.
        """
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            output_directory=output_directory,
            **kwargs,
        )
        self.population_sizes: list[int] = population_sizes
        self.mutation_rates: list[float] = mutation_rates
        self.hamming_factors: list[float] = hamming_factors
        self.hamming_factor_decays: list[float] = hamming_factor_decays

    def run(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Run the genetic algorithm experiment.

        This method performs grid search over the provided population sizes,
        mutation rates, and optional Hamming factors and decays, and returns
        the statistics and curves generated by the experiment.

        Returns
        -------
        tuple
            A tuple containing two DataFrames: run statistics and run curves
        """
        return super().run_experiment_(
            algorithm=genetic_alg,
            pop_size=("Population Size", self.population_sizes),
            mutation_prob=("Mutation Rate", self.mutation_rates),
            hamming_factor=("Hamming Factor", self.hamming_factors),
            hamming_decay_factor=("Hamming Factor Decay Rate", self.hamming_factor_decays),
        )



"""Class for defining an exponential decay schedule for Simulated Annealing (SA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

import numpy as np


class ExpDecay:
    """
    Defines an exponential decay schedule for the temperature parameter T in simulated annealing,
    using the formula:

    .. math::

        T(t) = \\max(T_{0} e^{-rt}, T_{min})

    where :math:`T_{0}` is the initial temperature at time `t = 0`, `r` is the rate of exponential decay,
    and :math:`T_{min}` is the minimum allowable temperature.

    Parameters
    ----------
    init_temp : float, default=1.0
        The initial value of the temperature parameter T. Must be greater than 0.
    exp_const : float, default=0.005
        The rate of exponential decay. Must be greater than 0.
    min_temp : float, default=0.001
        The minimum allowable temperature. Must be greater than 0 and less than `init_temp`.

    Attributes
    ----------
    init_temp : float
        Stores the initial temperature.
    exp_const : float
        Stores the rate of exponential decay.
    min_temp : float
        Stores the minimum temperature.

    Examples
    --------
    >>> schedule = ExpDecay(init_temp=10, exp_const=0.05, min_temp=1)
    >>> print(schedule.evaluate(5))
    7.788007830714049
    """

    def __init__(self, init_temp: float = 1.0, exp_const: float = 0.005, min_temp: float = 0.001):
        self.init_temp: float = init_temp
        self.exp_const: float = exp_const
        self.min_temp: float = min_temp

        if self.init_temp <= 0:
            raise ValueError("Initial temperature must be greater than 0.")
        if self.exp_const <= 0:
            raise ValueError("Decay rate must be greater than 0 and positive.")
        if not (0 < self.min_temp < self.init_temp):
            raise ValueError("Minimum temperature must be greater than 0 and less than initial temperature.")

    def __str__(self) -> str:
        return f"ExpDecay(init_temp={self.init_temp}, exp_const={self.exp_const}, min_temp={self.min_temp})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ExpDecay):
            return False
        return self.init_temp == other.init_temp and self.exp_const == other.exp_const and self.min_temp == other.min_temp

    def evaluate(self, t: int) -> float:
        """
        Evaluate the temperature parameter at the specified time using exponential decay.

        Parameters
        ----------
        t : int
            The time at which the temperature parameter T is evaluated.

        Returns
        -------
        float
            The temperature parameter at the given time, respecting the minimum temperature.
        """
        return float(max(self.init_temp * np.exp(-self.exp_const * t), self.min_temp))

    def get_info__(self, t: int = None, prefix: str = "") -> dict:
        """
        Retrieve a dictionary containing the configuration and optionally the current value of the decay schedule.

        Parameters
        ----------
        t : int | None, optional
            If provided, include the current temperature value at the given time.
        prefix : str, optional
            A prefix to append to each dictionary key, enhancing integration with other data structures.

        Returns
        -------
        dict
            A dictionary detailing the decay schedule settings and optionally the current temperature.
        """
        info_prefix = f"{prefix}schedule_" if prefix else "schedule_"

        info = {
            f"{info_prefix}type": "exponential",
            f"{info_prefix}init_temp": self.init_temp,
            f"{info_prefix}exp_const": self.exp_const,
            f"{info_prefix}min_temp": self.min_temp,
        }

        if t is not None:
            info[f"{info_prefix}current_value"] = self.evaluate(t)

        return info



"""Class for defining a geometric decay schedule for Simulated Annealing (SA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause


class GeomDecay:
    """
    Defines a geometric decay schedule for the temperature parameter T in simulated annealing,
    using the formula:

    .. math::

        T(t) = \\max(T_{0} \\times r^{t}, T_{min})

    where :math:`T_{0}` is the initial temperature at time `t = 0`, `r` is the rate of geometric decay,
    and :math:`T_{min}` is the minimum allowable temperature.

    Parameters
    ----------
    init_temp : float, default=1.0
        The initial value of the temperature parameter T. Must be greater than 0.
    decay : float, default=0.99
        The rate of geometric decay. Must be between 0 (exclusive) and 1 (inclusive).
    min_temp : float, default=0.001
        The minimum allowable temperature. Must be greater than 0 and less than `init_temp`.

    Attributes
    ----------
    init_temp : float
        Stores the initial temperature.
    decay : float
        Stores the rate of geometric decay.
    min_temp : float
        Stores the minimum temperature.

    Examples
    --------
    >>> schedule = GeomDecay(init_temp=10, decay=0.95, min_temp=1)
    >>> print(schedule.evaluate(5))
    7.737809374999998
    """

    def __init__(self, init_temp: float = 1.0, decay: float = 0.99, min_temp: float = 0.001):
        self.init_temp: float = init_temp
        self.decay: float = decay
        self.min_temp: float = min_temp

        if self.init_temp <= 0:
            raise ValueError("Initial temperature must be greater than 0.")
        if not (0 < self.decay <= 1):
            raise ValueError("Decay rate must be between 0 and 1, exclusive of 0.")
        if not (0 < self.min_temp < self.init_temp):
            raise ValueError("Minimum temperature must be greater than 0 and less than initial temperature.")

    def __str__(self):
        return str(self.init_temp)

    def __repr__(self):
        return f"{self.__class__.__name__}(init_temp={self.init_temp}, decay={self.decay}, min_temp={self.min_temp})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GeomDecay):
            return False
        return self.init_temp == other.init_temp and self.decay == other.decay and self.min_temp == other.min_temp

    def evaluate(self, t: int) -> float:
        """
        Evaluate the temperature parameter at the specified time using geometric decay.

        Parameters
        ----------
        t : int
            The time at which the temperature parameter T is evaluated.

        Returns
        -------
        float
            The temperature parameter at the given time, respecting the minimum temperature.
        """
        return float(max(self.init_temp * (self.decay**t), self.min_temp))

    def get_info__(self, t: int = None, prefix: str = "") -> dict:
        """
        Retrieve a dictionary containing the configuration and optionally the current value of the decay schedule.

        Parameters
        ----------
        t : int | None, optional
            If provided, include the current temperature value at the given time.
        prefix : str, optional
            A prefix to append to each dictionary key, enhancing integration with other data structures.

        Returns
        -------
        dict
            A dictionary detailing the decay schedule settings and optionally the current temperature.
        """
        full_prefix = f"{prefix}__schedule_" if len(prefix) else "schedule_"

        info = {
            f"{full_prefix}type": "geometric",
            f"{full_prefix}init_temp": self.init_temp,
            f"{full_prefix}decay": self.decay,
            f"{full_prefix}min_temp": self.min_temp,
        }

        if t is not None:
            info[f"{full_prefix}current_value"] = self.evaluate(t)

        return info



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np

from mlrose_ky.algorithms.crossovers import OnePointCrossOver
from mlrose_ky.algorithms.mutators import ChangeOneMutator
from mlrose_ky.fitness import FlipFlop
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class FlipFlopOpt(DiscreteOpt):
    """Class for defining FlipFlop optimization problems.

    Parameters
    ----------
    length : int, default=None
        Number of elements in the state vector. If not specified, it is inferred from the fitness function.

    fitness_fn : Any, default=None
        Object to implement the fitness function for optimization. If not specified, defaults to the `FlipFlop` fitness function.

    maximize : bool, default=True
        Whether to maximize the fitness function. Set :code:`False` for minimization problems.

    crossover : OnePointCrossOver, default=None
        Crossover operation used for reproduction. If None, defaults to `OnePointCrossOver`.

    mutator : ChangeOneMutator, default=None
        Mutation operation used for reproduction. If None, defaults to `ChangeOneMutator`.

    Attributes
    ----------
    length : int
        Number of elements in the state vector.

    fitness_fn : FlipFlop
        Fitness function for the optimization problem.

    population : np.ndarray
        Array containing the current population.

    pop_fitness : np.ndarray
        Array containing the fitness values for the current population.

    max_val : int
        Number of unique values that each element in the state vector can take (always 2 for FlipFlopOpt).
    """

    def __init__(
        self,
        length: int = None,
        fitness_fn: Any = None,
        maximize: bool = True,
        crossover: "OnePointCrossOver" = None,
        mutator: "ChangeOneMutator" = None,
    ):
        if (fitness_fn is None) and (length is None):
            raise ValueError("fitness_fn or length must be specified.")

        if length is None:
            length = len(fitness_fn.weights)

        self.length: int = length

        if fitness_fn is None:
            fitness_fn = FlipFlop()

        self.max_val: int = 2
        crossover = OnePointCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator

        super().__init__(length, fitness_fn, maximize, crossover=crossover, mutator=mutator)

        # Set initial state
        state = np.random.randint(2, size=self.length)
        self.set_state(state)

    def evaluate_population_fitness(self):
        """Calculate fitness for the current population."""
        self.pop_fitness = self.fitness_fn.evaluate_many(self.population)

    def random_pop(self, pop_size: int):
        """Create a population of random state vectors.

        Parameters
        ----------
        pop_size : int
            Size of population to be created.

        Raises
        ------
        ValueError
            If pop_size is not a positive integer.
        """
        if pop_size <= 0:
            raise ValueError("pop_size must be a positive integer.")
        elif not isinstance(pop_size, int):
            if pop_size.is_integer():
                pop_size = int(pop_size)
            else:
                raise ValueError("pop_size must be a positive integer.")

        # Generate random population
        population = np.random.rand(pop_size, self.length)
        population[population < 0.5] = 0
        population[population >= 0.5] = 1
        self.population = population

        # Evaluate fitness for the population
        self.evaluate_population_fitness()

    def can_stop(self) -> bool:
        """Determine if the optimization process can stop.

        Returns
        -------
        bool
            True if the fitness equals the length of the state vector minus 1, otherwise False.
        """
        return int(self.get_fitness()) == int(self.length - 1)



"""One Point Crossover implementation for Genetic Algorithms (GA).

This module defines a one-point crossover operation used in genetic algorithms,
where a single crossover point is chosen randomly to combine two parent solutions.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any, Sequence

import numpy as np

from mlrose_ky.algorithms.crossovers._crossover_base import _CrossOverBase


class OnePointCrossOver(_CrossOverBase):
    """
    One-point crossover for genetic algorithms.

    This class implements a simple one-point crossover, where a single crossover
    point on the parent chromosomes is chosen randomly, and the genetic information
    is exchanged to create a new offspring.

    Inherits from:
    _CrossOverBase : Abstract base class for crossover operations.
    """

    def __init__(self, opt_prob: Any):
        """
        Initialize the OnePointCrossOver with the given optimization problem.

        Parameters
        ----------
        opt_prob : Any
            An instance of the optimization problem related to the genetic algorithm.
        """
        super().__init__(opt_prob)

    def mate(self, p1: Sequence[float], p2: Sequence[float]) -> np.ndarray:
        """
        Perform the one-point crossover between two parent sequences to produce offspring.

        A single crossover point is selected randomly from the chromosome. All genes before
        that point are copied from the first parent and the rest from the second parent.

        Parameters
        ----------
        p1 : Sequence[float]
            The first parent chromosome sequence.
        p2 : Sequence[float]
            The second parent chromosome sequence.

        Returns
        -------
        np.ndarray
            The offspring chromosome resulting from the crossover.
        """
        crossover_point = 1 + np.random.randint(self._length - 1)
        return np.array([*p1[:crossover_point], *p2[crossover_point:]])



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from sklearn.base import RegressorMixin

from mlrose_ky.algorithms.decay import GeomDecay
from ._nn_core import _NNCore


class LinearRegression(_NNCore, RegressorMixin):
    """
    Class for defining linear regression weights optimization problem.
    Inherits :code:`fit` and :code:`predict` methods from the
    :code:`NeuralNetwork()` class.

    Parameters
    ----------
    algorithm : str, default='random_hill_climb'
        Algorithm used to find optimal network weights. Must be one of:
        'random_hill_climb', 'simulated_annealing', 'genetic_alg', or
        'gradient_descent'.

    max_iters : int, default=100
        Maximum number of iterations used to fit the weights.

    bias : bool, default=True
        Whether to include a bias term.

    learning_rate : float, default=0.1
        Learning rate for gradient descent or step size for randomized
        optimization algorithms.

    early_stopping : bool, default=False
        Whether to terminate the algorithm early if the loss is not improving.
        If :code:`True`, then stop after max_attempts iterations with no
        improvement.

    clip_max : float, default=1e+10
        Used to limit weights to the range [-1 * clip_max, clip_max].

    restarts : int, default=0
        Number of random restarts. Only required if :code:`algorithm='random_hill_climb'`.

    schedule : GeomDecay, default=mlrose_ky.GeomDecay()
        Schedule used to determine the value of the temperature parameter.
        Only required if :code:`algorithm='simulated_annealing'`.

    pop_size : int, default=200
        Size of population. Only required if :code:`algorithm='genetic_alg'`.

    mutation_prob : float, default=0.1
        Probability of a mutation at each element of the state vector during
        reproduction, expressed as a value between 0 and 1. Only required if
        :code:`algorithm='genetic_alg'`.

    max_attempts : int, default=10
        Maximum number of attempts to find a better state. Only required if
        :code:`early_stopping=True`.

    random_state : int or None, default=None
        If random_state is a positive integer, it is used as the seed by
        np.random.seed(); otherwise, the random seed is not set.

    curve : bool, default=False
        If True, the curve containing the fitness at each training iteration
        is returned.

    Attributes
    ----------
    fitted_weights : np.ndarray
        Numpy array giving the fitted weights after :code:`fit`.

    loss : float
        Value of the loss function for fitted weights after :code:`fit`.

    fitness_curve : np.ndarray
        Numpy array giving the fitness at each training iteration.
    """

    def __init__(
        self,
        algorithm: str = "random_hill_climb",
        max_iters: int = 100,
        bias: bool = True,
        learning_rate: float = 0.1,
        early_stopping: bool = False,
        clip_max: float = 1e10,
        restarts: int = 0,
        schedule: GeomDecay = GeomDecay(),
        pop_size: int = 200,
        mutation_prob: float = 0.1,
        max_attempts: int = 10,
        random_state: int = None,
        curve: bool = False,
    ):
        # Initialize LinearRegression with neural network configurations
        super().__init__(
            hidden_nodes=[],
            activation="identity",
            algorithm=algorithm,
            max_iters=max_iters,
            bias=bias,
            is_classifier=False,
            learning_rate=learning_rate,
            early_stopping=early_stopping,
            clip_max=clip_max,
            restarts=restarts,
            schedule=schedule,
            pop_size=pop_size,
            mutation_prob=mutation_prob,
            max_attempts=max_attempts,
            random_state=random_state,
            curve=curve,
        )



"""Base class for running neural network weights optimization experiments, including grid search functionality."""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

import hashlib
import logging
import os
import pickle as pk
import time
from abc import ABC
from typing import Callable, Any

import numpy as np
import pandas as pd

from mlrose_ky.gridsearch import GridSearchMixin
from mlrose_ky.runners._runner_base import _RunnerBase


class _NNRunnerBase(_RunnerBase, GridSearchMixin, ABC):
    """
    A base class for running neural network experiments with grid search.
    It extends functionality from _RunnerBase and GridSearchMixin.

    This class provides methods for setting up and executing grid search over neural network hyperparameters,
    handling cross-validation, saving results, and managing file operations.

    Attributes
    ----------
    x_train : np.ndarray
        Features for the training data.
    y_train : np.ndarray
        Labels for the training data.
    x_test : np.ndarray
        Features for the testing data.
    y_test : np.ndarray
        Labels for the testing data.
    _experiment_name : str
        Name of the experiment.
    seed : int
        Random seed to ensure reproducibility.
    iteration_list : np.ndarray | list[int]
        List of iteration counts to perform.
    grid_search_parameters : dict
        Hyperparameters for grid search.
    cv : int
        Number of cross-validation folds.
    generate_curves : bool
        Whether to generate learning curves during the experiment.
    _output_directory : str | None
        Directory where outputs will be saved.
    verbose_grid_search : bool
        Whether to output detailed grid search information.
    override_ctrl_c_handler : bool
        Whether to override the default CTRL+C handler.
    n_jobs : int
        Number of parallel jobs for grid search.
    cv_results_df : pd.DataFrame | None
        DataFrame to store cross-validation results.
    best_params : dict | None
        Dictionary storing the best parameters found during grid search.
    """

    _interrupted_results: list = []

    def __init__(
        self,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_test: np.ndarray,
        y_test: np.ndarray,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        grid_search_parameters: dict,
        grid_search_scorer_method: Callable,
        cv: int = 5,
        generate_curves: bool = True,
        output_directory: str = None,
        verbose_grid_search: bool = True,
        override_ctrl_c_handler: bool = True,
        n_jobs: int = 1,
        replay: bool = False,
        **kwargs,
    ):
        """
        Initializes the _NNRunnerBase class.

        Parameters
        ----------
        x_train : np.ndarray
            Features for the training data.
        y_train : np.ndarray
            Labels for the training data.
        x_test : np.ndarray
            Features for the testing data.
        y_test : np.ndarray
            Labels for the testing data.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed to ensure reproducibility.
        iteration_list : np.ndarray | list[int]
            List of iteration counts to perform.
        grid_search_parameters : dict
            Hyperparameters for grid search.
        grid_search_scorer_method : Callable
            Scorer method for evaluating the grid search.
        cv : int, optional, default=5
            Number of cross-validation folds.
        generate_curves : bool, optional, default=True
            Whether to generate learning curves.
        output_directory : str | None, optional, default=None
            Directory where outputs will be saved.
        verbose_grid_search : bool, optional, default=True
            Whether to output detailed grid search information.
        override_ctrl_c_handler : bool, optional, default=True
            Whether to override the default CTRL+C handler.
        n_jobs : int, optional, default=1
            Number of parallel jobs for grid search.
        replay : bool, optional, default=False
            Whether to replay previous results.
        **kwargs :
            Additional hyperparameters for grid search.
        """
        super().__init__(
            problem=None,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            generate_curves=generate_curves,
            output_directory=output_directory,
            replay=replay,
            override_ctrl_c_handler=override_ctrl_c_handler,
            copy_zero_curve_fitness_from_first=True,
        )

        GridSearchMixin.__init__(self, scorer_method=grid_search_scorer_method)

        self.classifier: Any = None
        self.grid_search_parameters: dict = self.build_grid_search_parameters(grid_search_parameters, **kwargs)
        self.x_train: np.ndarray = x_train
        self.y_train: np.ndarray = y_train
        self.x_test: np.ndarray = x_test
        self.y_test: np.ndarray = y_test
        self.cv: int = cv
        self.n_jobs: int = n_jobs
        self.verbose_grid_search: bool = verbose_grid_search
        self.cv_results_df: pd.DataFrame | None = None
        self.best_params: dict | None = None

    def run(self) -> tuple[pd.DataFrame | None, pd.DataFrame | None, pd.DataFrame | None, Any | None]:
        """
        Executes the runner, performing grid search and handling the results.

        Returns
        -------
        tuple
            Contains the run statistics DataFrame, curves DataFrame, CV results DataFrame, and the grid search result.
        """
        try:
            self._setup()

            logging.info(f"Running experiment: {self._experiment_name}")

            # Replay mode allows reusing previous grid search results, avoiding re-execution
            if self.replay_mode():
                gsr_name = f"{super()._get_pickle_filename_root('grid_search_results')}.p"
                with open(gsr_name, "rb") as pickle_file:
                    search_results = pk.load(pickle_file)
            else:
                run_start = time.perf_counter()
                search_results = self._perform_grid_search(
                    classifier=self.classifier,
                    parameters=self.grid_search_parameters,
                    x_train=self.x_train,
                    y_train=self.y_train,
                    cv=self.cv,
                    n_jobs=self.n_jobs,
                    verbose=self.verbose_grid_search,
                )
                run_end = time.perf_counter()
                logging.info(f"Run time: {run_end - run_start}")

            # Updates internal attributes with the best estimator found by grid search
            self.__dict__.update(search_results.best_estimator_.runner.__dict__)
            self.best_params = search_results.best_params_

            self.cv_results_df = self._make_cv_results_data_frame(search_results.cv_results_)
            extra_data_frames = {"cv_results_df": self.cv_results_df}

            self._create_and_save_run_data_frames(extra_data_frames=extra_data_frames, final_save=True)

            # Save grid search results to disk
            try:
                self._dump_pickle_to_disk(search_results, "grid_search_results", final_save=True)
            except:
                pass

            # Perform predictions and score using the best estimator found in the search
            try:
                y_pred = search_results.best_estimator_.predict(self.x_test)
                score = self.score(y_pred=y_pred, y_true=self.y_train)
                self._print_banner(f"Score: {score}")
            except:
                pass

            return self.run_stats_df, self.curves_df, self.cv_results_df, search_results
        except KeyboardInterrupt:
            # Handling graceful termination in case of manual interruption
            return None, None, None, None
        finally:
            self._tear_down()

    def _get_pickle_filename_root(self, name: str) -> str:
        """
        Generates a root filename for pickle files with a hash based on the current algorithm arguments.

        Parameters
        ----------
        name : str
            The base name for the pickle file.

        Returns
        -------
        str
            The root filename with a hash appended.
        """
        filename_root = super()._get_pickle_filename_root(name)

        # Create a unique hash based on argument values (excluding state-related keys)
        arg_text = "".join([f"{k}_{self._sanitize_value(v)}_" for k, v in self._current_logged_algorithm_args.items() if "state" not in k])
        arg_hash = f"__{hashlib.md5(arg_text.encode()).hexdigest()}".upper() if len(arg_text) > 0 else ""

        return filename_root + arg_hash

    @staticmethod
    def _check_match(df_reference: pd.DataFrame, df_to_check: pd.DataFrame) -> bool:
        """
        Checks if two DataFrames have matching rows based on their columns.

        Parameters
        ----------
        df_reference : pd.DataFrame
            The reference DataFrame to check against.
        df_to_check : pd.DataFrame
            The DataFrame to check for matching rows.

        Returns
        -------
        bool
            True if a matching row is found, False otherwise.
        """
        cols = [col for col in df_reference.columns]

        for _, row in df_to_check.iterrows():
            # Returns True as soon as a match is found
            if all(df_reference[col][0] == row[col] for col in cols):
                return True

        return False

    def _tear_down(self, filename: str | None = None):
        """
        Finalizes the runner, ensuring that the proper files are saved or cleaned up.

        Parameters
        ----------
        filename : str | None, optional, default=None
            Filename to clean up.
        """
        if self.best_params is None or self.replay_mode() is None or self._output_directory is None:
            super()._tear_down()
            return

        filename_root = super()._get_pickle_filename_root("")
        path = os.path.join(*filename_root.split(os.sep)[:-1])
        filename_part = filename_root.split(os.sep)[-1]

        if not os.path.isdir(path) and path[0] != os.sep:
            path = f"{os.sep}{path}"

        filenames = [fn for fn in os.listdir(str(path)) if (filename_part in fn and fn.endswith(".p") and "_df_" in fn)]

        if not filenames:
            raise FileNotFoundError(f"No matching filenames found in path: {path}")

        df_best_params = pd.DataFrame([{k: self._sanitize_value(v) for k, v in self.best_params.items()}])

        correct_files = []
        incorrect_files = []
        for fn in filenames:
            filename = os.path.join(str(path), fn)
            with open(filename, "rb") as pickle_file:
                try:
                    df = pk.load(pickle_file)
                    found = self._check_match(df_best_params, df)
                    (correct_files if found else incorrect_files).append(filename)
                except (EOFError, pk.PickleError):
                    pass

        # Extracts md5 hashes from correct and incorrect files for renaming or deletion
        correct_md5s = {p.split("_")[-1][:-2] for p in correct_files}
        incorrect_md5s = {p.split("_")[-1][:-2] for p in incorrect_files}

        # Remove files corresponding to incorrect md5 hashes
        all_incorrect_files = [
            os.path.join(str(path), fn) for incorrect_md5 in incorrect_md5s for fn in os.listdir(str(path)) if incorrect_md5 in fn
        ]
        for _filename in all_incorrect_files:
            os.remove(_filename)

        # Rename correct files by removing md5 hash from the filename
        all_correct_files = [
            (os.path.join(str(path), fn), f"__{_correct_md5}")
            for _correct_md5 in correct_md5s
            for fn in os.listdir(str(path))
            if _correct_md5 in fn
        ]
        for _filename, _correct_md5 in all_correct_files:
            correct_filename = _filename.replace(_correct_md5, "")
            if os.path.exists(correct_filename):
                os.rename(correct_filename, f"{correct_filename}.bak")
            os.rename(_filename, correct_filename)

        super()._tear_down()

    @staticmethod
    def _make_cv_results_data_frame(cv_results: dict) -> pd.DataFrame:
        """
        Creates a DataFrame from cross-validation results.

        Parameters
        ----------
        cv_results : dict
            Cross-validation results.

        Returns
        -------
        pd.DataFrame
            The cross-validation results as a DataFrame.
        """
        cv_results = cv_results.copy()
        param_prefix = "param_"
        param_labels = [k for k in cv_results if param_prefix in k]

        new_param_values = {p: [] for p in param_labels}
        for v in cv_results["params"]:
            for p in param_labels:
                param_label = p.replace(param_prefix, "")
                new_param_values[p].append(_NNRunnerBase._sanitize_value(v[param_label]))

        cv_results.update(new_param_values)
        df = pd.DataFrame(cv_results)
        df.dropna(inplace=True)

        return df

    @staticmethod
    def build_grid_search_parameters(grid_search_parameters: dict, **kwargs) -> dict:
        """
        Builds a dictionary of grid search parameters by combining initial parameters and additional arguments.

        Parameters
        ----------
        grid_search_parameters : dict
            Initial grid search parameters.
        **kwargs :
            Additional hyperparameters for grid search.

        Returns
        -------
        dict
            Combined grid search parameters.
        """
        return {**grid_search_parameters, **kwargs}

    def _grid_search_score_intercept(
        self, y_true: np.ndarray, y_pred: np.ndarray, sample_weight: np.ndarray = None, adjusted: bool = False
    ) -> float:
        """
        Intercepts the grid search scoring process to handle special cases, particularly in aborted runs.

        Parameters
        ----------
        y_true : np.ndarray
            True labels.
        y_pred : np.ndarray
            Predicted labels.
        sample_weight : np.ndarray | None, optional, default=None
            Sample weights for scoring.
        adjusted : bool, optional, default=False
            Whether to adjust the score.

        Returns
        -------
        float
            The grid search score, or NaN if the run was aborted.
        """
        if not self.classifier.fit_started_ and self.has_aborted():
            return np.NaN

        return super()._grid_search_score_intercept(y_true=y_true, y_pred=y_pred, sample_weight=sample_weight, adjusted=adjusted)



"""
Class for running optimization experiments using mlrose_ky.NNClassifier, including grid search functionality.

Example usage:
    from mlrose_ky.runners import NNGSRunner

    grid_search_parameters = ({
        'max_iters': [1, 2, 4, 8, 16, 32, 64, 128],                     # nn params
        'learning_rate': [0.001, 0.002, 0.003],                         # nn params
        'schedule': [ArithDecay(1), ArithDecay(100), ArithDecay(1000)]  # sa params
    })

    nnr = NNGSRunner(x_train=x_train,
                     y_train=y_train,
                     x_test=x_test,
                     y_test=y_test,
                     experiment_name='nn_test',
                     algorithm=mlrose_ky.algorithms.sa.simulated_annealing,
                     grid_search_parameters=grid_search_parameters,
                     iteration_list=[1, 10, 50, 100, 250, 500, 1000, 2500, 5000, 10000],
                     hidden_layer_sizes=[[44,44]],
                     bias=True,
                     early_stopping=False,
                     clip_max=1e+10,
                     max_attempts=500,
                     generate_curves=True,
                     seed=200972)

    results: skms.GridSearchCV = nnr.run()
"""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

from typing import Any, Callable, Optional

import numpy as np
import sklearn.metrics as skmt

from mlrose_ky.decorators import short_name, get_short_name
from mlrose_ky.neural import NNClassifier
from mlrose_ky.runners._nn_runner_base import _NNRunnerBase


@short_name("nngs")
class NNGSRunner(_NNRunnerBase):
    """
    A runner for performing optimization experiments using the mlrose_ky.NNClassifier.

    This class extends _NNRunnerBase and provides grid search functionality for optimizing
    the hyperparameters of an NNClassifier model.

    Attributes
    ----------
    classifier : NNClassifier
        The classifier used for the experiment.
    """

    def __init__(
        self,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_test: np.ndarray,
        y_test: np.ndarray,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        algorithm: Any,
        grid_search_parameters: dict,
        grid_search_scorer_method: Optional[Callable] = skmt.balanced_accuracy_score,
        bias: bool = True,
        early_stopping: bool = True,
        clip_max: float = 1e10,
        max_attempts: int = 500,
        n_jobs: int = 1,
        cv: int = 5,
        generate_curves: bool = True,
        output_directory: str = None,
        **kwargs: Any,
    ):
        """
        Initialize the NNGSRunner class with training and testing data and various experiment parameters.

        Parameters
        ----------
        x_train : np.ndarray
            Training input data.
        y_train : np.ndarray
            Target labels for training data.
        x_test : np.ndarray
            Test input data.
        y_test : np.ndarray
            Target labels for test data.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed for reproducibility.
        iteration_list : np.ndarray | list of int
            List of iterations for the experiment.
        algorithm: Any
            The optimization algorithm to be used (e.g., simulated_annealing).
        grid_search_parameters : dict
            Parameters for grid search.
        grid_search_scorer_method : callable, optional
            Scoring method for grid search.
        bias : bool, optional
            Whether to use bias in the NNClassifier.
        early_stopping : bool, optional
            Whether to stop early if no improvement is detected.
        clip_max : float, optional
            Maximum value for gradient clipping.
        max_attempts : int, optional
            Maximum number of attempts without improvement before stopping.
        n_jobs : int, optional
            Number of jobs to run in parallel.
        cv : int, optional
            Number of cross-validation folds.
        generate_curves : bool, optional
            Whether to generate learning curves.
        output_directory : str, optional
            Directory to save output.
        """
        # Take a copy of the grid search parameters
        grid_search_parameters = {**grid_search_parameters}

        # Hack for compatibility purposes
        if "max_iter" in grid_search_parameters:
            grid_search_parameters["max_iter"] = grid_search_parameters.pop("max_iters")

        # Call the base class init
        super().__init__(
            x_train=x_train,
            y_train=y_train,
            x_test=x_test,
            y_test=y_test,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            grid_search_parameters=grid_search_parameters,
            generate_curves=generate_curves,
            output_directory=output_directory,
            n_jobs=n_jobs,
            cv=cv,
            grid_search_scorer_method=grid_search_scorer_method,
            **kwargs,
        )

        # Build the classifier
        self.classifier = NNClassifier(
            runner=self,
            algorithm=algorithm,
            max_attempts=max_attempts,
            clip_max=clip_max,
            early_stopping=early_stopping,
            seed=seed,
            bias=bias,
        )

        # Update short name based on the algorithm
        self._set_dynamic_runner_name(f"{get_short_name(self)}_{get_short_name(algorithm)}")

    def run_one_experiment_(self, algorithm: Any, total_args: dict, **params: dict) -> tuple | None:
        """
        Run one instance of the experiment with the specified algorithm and parameters.

        Parameters
        ----------
        algorithm: Any
            The optimization algorithm to run.
        total_args : dict
            Dictionary of arguments to pass to the algorithm.
        **params : dict
            Additional parameters for the experiment.

        Returns
        -------
        tuple | None
            The results of the experiment.
        """
        if self._extra_args is not None and len(self._extra_args) > 0:
            params = {**params, **self._extra_args}

        total_args.update(params)
        total_args.pop("problem")
        user_info = [(k, v) for k, v in total_args.items()]

        return self._invoke_algorithm(
            algorithm=algorithm, curve=self.generate_curves, user_info=user_info, additional_algorithm_args=total_args, **params
        )



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import networkx as nx
import numpy as np

from mlrose_ky.algorithms.crossovers import UniformCrossOver
from mlrose_ky.algorithms.mutators import ChangeOneMutator
from mlrose_ky.fitness import MaxKColor
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class MaxKColorOpt(DiscreteOpt):
    """Class for defining Max-K Color optimization problems.

    Parameters
    ----------
    edges : list[tuple[int, int]], default=None
        List of edges in the graph. Each edge is represented as a tuple of two node indices.

    length : int, default=None
        Number of elements in the state vector. If not specified, it is inferred from the edges or fitness function.

    fitness_fn : MaxKColor, default=None
        Object to implement the fitness function for optimization. If not provided, a new `MaxKColor` function will be created.

    maximize : bool, default=False
        Whether to maximize the fitness function. Set :code:`False` for minimization problems.

    max_colors : int, default=None
        Maximum number of colors to use for coloring the graph.

    crossover : UniformCrossOver, default=None
        Crossover operation used for reproduction. If None, defaults to `UniformCrossOver`.

    mutator : ChangeOneMutator, default=None
        Mutation operation used for reproduction. If None, defaults to `ChangeOneMutator`.

    source_graph : networkx.Graph, default=None
        Graph object representing the graph to be colored. If not provided, a new graph is created based on the edges.

    Attributes
    ----------
    length : int
        Number of elements in the state vector.

    fitness_fn : MaxKColor
        Fitness function for the optimization problem.

    max_val : int
        Maximum number of colors used in the state vector.

    source_graph : nx.Graph
        Graph representing the nodes and edges for the coloring problem.

    stop_fitness : int
        The fitness value at which the optimization process can stop.
    """

    def __init__(
        self,
        edges: list[tuple[int, int]] = None,
        length: int = None,
        fitness_fn: Any = None,
        maximize: bool = False,
        max_colors: int = None,
        crossover: "UniformCrossOver" = None,
        mutator: "ChangeOneMutator" = None,
        source_graph: nx.Graph = None,
    ):
        # Ensure that either fitness_fn or edges are provided
        if fitness_fn is None and edges is None:
            raise ValueError("Either fitness_fn or edges must be specified.")

        # If length is not provided, infer it from the edges or the fitness function
        if length is None:
            if fitness_fn is None:
                length = len(edges)  # Infer from the number of edges
            else:
                length = len(fitness_fn.weights)  # Infer from the fitness function

        self.length: int = length

        # If fitness function is not provided, create a MaxKColor fitness function
        if fitness_fn is None:
            fitness_fn = MaxKColor(edges, maximize)

        # Special handling for a single-node graph
        if length == 1:
            self.max_val: int = 1
            self.source_graph: nx.Graph = nx.Graph()  # Create an empty graph
            fitness_fn.set_graph(self.source_graph)
            self.stop_fitness: int = 0 if maximize else 0
            super().__init__(length, fitness_fn, maximize, 1, crossover, mutator)
            self.set_state(np.array([0]))
            return

        # Create or update the graph based on the provided edges or source_graph
        if source_graph is None:
            g = nx.Graph()
            g.add_edges_from(edges)
            # Ensure all nodes up to `length` are added to the graph
            for i in range(length):
                if i not in g:
                    g.add_node(i)
            self.source_graph = g
        else:
            self.source_graph = source_graph

        # Set stop fitness based on whether the problem is maximization or minimization
        self.stop_fitness: int = self.source_graph.number_of_edges() if maximize else 0

        fitness_fn.set_graph(self.source_graph)

        # Automatically calculate max_colors if not provided
        if max_colors is None:
            total_neighbor_count = [len(list(self.source_graph.neighbors(n))) for n in range(length)]
            if total_neighbor_count:
                max_colors = 1 + max(total_neighbor_count)  # Set to 1 + max neighbors
            else:
                max_colors = 1  # Default to 1 when there are no edges
        self.max_val: int = max_colors

        # Use default crossover and mutator if none are provided
        crossover = UniformCrossOver(self) if crossover is None else crossover
        mutator = ChangeOneMutator(self) if mutator is None else mutator
        super().__init__(length, fitness_fn, maximize, max_colors, crossover, mutator)

        # Initialize the state with a shuffled random assignment of colors
        state = np.random.randint(max_colors, size=self.length)
        np.random.shuffle(state)
        self.set_state(state)

    def can_stop(self) -> bool:
        """Determine if the optimization process can stop.

        Returns
        -------
        bool
            True if the fitness value matches the stopping criterion, otherwise False.
        """
        return int(self.get_fitness()) == self.stop_fitness



"""
Class for running optimization experiments using Simulated Annealing (SA), including grid search functionality.

Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    # note that you can also initialize a temperature_list this way
    # temperature_list = [mlrose_ky.GeomDecay(init_temp=t, decay=d) for (t, d) in [(1, 0.99), (1e2, 0.999)]]
    # if you use this form, the decay_list parameter is ignored.

    sa = SARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(14),
                  max_attempts=5000,
                  temperature_list=[1, 10, 50, 100, 250, 500, 1000, 2500, 5000, 10000],
                  decay_list=[mlrose_ky.GeomDecay])

    df_run_stats, df_run_curves = sa.run()
"""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

from typing import Any, Callable

import numpy as np
import pandas as pd

from mlrose_ky.algorithms import simulated_annealing
from mlrose_ky.algorithms.decay import GeomDecay
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase


@short_name("sa")
class SARunner(_RunnerBase):
    """
    A runner for performing optimization experiments using Simulated Annealing (SA).

    This class extends _RunnerBase and provides functionality for running experiments
    with the SA algorithm, including grid search over hyperparameters such as
    temperature and decay schedules.

    Attributes
    ----------
    temperature_list : list[float]
        List of temperatures to test in the grid search.
    decay_list : list[Callable], optional
        List of decay schedules to use if raw temperature values are not provided.
    use_raw_temp : bool
        Whether to use raw temperature values or decay schedules.
    """

    def __init__(
        self,
        problem: Any,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        temperature_list: list[float],
        decay_list: list[Callable] | None = None,
        max_attempts: int = 500,
        generate_curves: bool = True,
        output_directory: str = None,
        **kwargs: dict,
    ):
        """
        Initialize the SARunner class with problem data and various experiment parameters.

        Parameters
        ----------
        problem : Any
            The optimization problem to be solved.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed for reproducibility.
        iteration_list : np.ndarray | list of int
            List of iterations for the experiment.
        temperature_list : list of float
            List of temperature values to test in the grid search.
        decay_list : list of callable, optional
            List of decay schedules to use, if raw temperature values are not provided.
        max_attempts : int, optional
            Maximum number of attempts without improvement before stopping.
        generate_curves : bool, optional
            Whether to generate learning curves.
        output_directory : str, optional
            Directory to save experiment result, default=None.
        """
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            output_directory=output_directory,
            **kwargs,
        )
        self.use_raw_temp = True
        self.temperature_list: list[float] = temperature_list

        # Use decay schedules if provided
        if all([np.isscalar(x) for x in temperature_list]):
            if decay_list is None:
                decay_list = [GeomDecay]

            self.decay_list: list[Callable] = decay_list
            self.use_raw_temp = False

    def run(self) -> tuple[pd.DataFrame | None, pd.DataFrame | None]:
        """
        Run the Simulated Annealing (SA) experiment.

        This method performs grid search over the provided temperature values and
        decay schedules, and returns the statistics and curves generated by the experiment.

        Returns
        -------
        tuple
            A tuple containing two DataFrames: run statistics and run curves.
        """
        temperatures = (
            self.temperature_list if self.use_raw_temp else [d(init_temp=t) for t in self.temperature_list for d in self.decay_list]
        )
        return super().run_experiment_(algorithm=simulated_annealing, schedule=("Temperature", temperatures))



"""Classes for defining algorithms problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from .ga import genetic_alg
from .gd import gradient_descent
from .hc import hill_climb
from .mimic import mimic
from .rhc import random_hill_climb
from .sa import simulated_annealing

from .crossovers import UniformCrossOver, TSPCrossOver, OnePointCrossOver

from .decay import ArithDecay, CustomSchedule, ExpDecay, GeomDecay

from .mutators import ChangeOneMutator, DiscreteMutator, ShiftOneMutator, SwapMutator



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import List, Tuple

import numpy as np


def flatten_weights(weights: List[np.ndarray]) -> np.ndarray:
    """
    Flatten list of weights arrays into a 1D array.

    Parameters
    ----------
    weights : list of np.ndarray
        List of 2D arrays for flattening.

    Returns
    -------
    np.ndarray
        1D array of flattened weights.
    """
    flat_weights = []

    for weight in weights:
        flat_weights += list(weight.flatten())

    return np.array(flat_weights)


def unflatten_weights(flat_weights: np.ndarray, node_list: List[int]) -> List[np.ndarray]:
    """
    Convert 1D weights array into list of 2D arrays.

    Parameters
    ----------
    flat_weights : np.ndarray
        1D weights array.

    node_list : list of int
        List giving the number of nodes in each layer of the network,
        including the input and output layers.

    Returns
    -------
    list of np.ndarray
        List of 2D arrays created from flat_weights.
    """
    nodes = sum(node_list[i] * node_list[i + 1] for i in range(len(node_list) - 1))

    if len(flat_weights) != nodes:
        raise ValueError(f"flat_weights must have length {nodes}, but got {len(flat_weights)}.")

    weights = []
    start = 0

    for i in range(len(node_list) - 1):
        end = start + node_list[i] * node_list[i + 1]
        weights.append(np.reshape(flat_weights[start:end], [node_list[i], node_list[i + 1]]))
        start = end

    return weights


def gradient_descent_original(
    problem,
    max_attempts: int = 10,
    max_iters: int | float = np.inf,
    init_state: np.ndarray = None,
    curve: bool = False,
    random_state: int = None,
) -> Tuple[np.ndarray, float, np.ndarray | None]:
    """
    Use gradient descent to find the optimal neural network weights.

    Parameters
    ----------
    problem : optimization object
        Object containing the optimization problem to be solved.
    max_attempts : int, default=10
        Maximum number of attempts to find a better state at each step.
    max_iters : int or float, default=np.inf
        Maximum number of iterations of the algorithm.
    init_state : np.ndarray, default=None
        Numpy array containing the starting state for the algorithm.
        If None, then a random state is used.
    curve : bool, default=False
        If True, returns a history of fitness values.
    random_state : int, default=None
        If provided, sets the random seed for reproducibility.

    Returns
    -------
    best_state : np.ndarray
        Numpy array containing the state that optimizes the fitness function.
    best_fitness : float
        Value of the fitness function at the best state.
    fitness_curve : np.ndarray or None
        Array containing the fitness at every iteration (if curve=True).
    """
    if not isinstance(max_attempts, int) or max_attempts < 0:
        raise ValueError(f"max_attempts must be a positive integer, got {max_attempts}.")

    if (not isinstance(max_iters, int) and max_iters != np.inf) or max_iters < 0:
        raise ValueError(f"max_iters must be a positive integer, got {max_iters}.")

    if init_state is not None and len(init_state) != problem.get_length():
        raise ValueError(f"init_state must have the same length as the problem, got {len(init_state)}.")

    # Set random seed
    if isinstance(random_state, int) and random_state > 0:
        np.random.seed(random_state)

    # Initialize problem
    if init_state is None:
        problem.reset()
    else:
        problem.set_state(init_state)

    fitness_curve = []
    attempts = 0
    iters = 0

    best_fitness = problem.get_maximize() * problem.get_fitness()
    best_state = problem.get_state()

    while attempts < max_attempts and iters < max_iters:
        iters += 1

        # Update weights
        updates = flatten_weights(problem.calculate_updates())
        next_state = problem.update_state(updates)
        next_fitness = problem.eval_fitness(next_state)

        if next_fitness > problem.get_fitness():
            attempts = 0
        else:
            attempts += 1

        if next_fitness > problem.get_maximize() * best_fitness:
            best_fitness = problem.get_maximize() * next_fitness
            best_state = next_state

        if curve:
            fitness_curve.append(problem.get_adjusted_fitness())

        problem.set_state(next_state)

    if curve:
        return best_state, best_fitness, np.asarray(fitness_curve)

    return best_state, best_fitness, None



"""
Class for running optimization experiments using MIMIC, including grid search functionality.

Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    mmc = MIMICRunner(problem=problem,
                      experiment_name=experiment_name,
                      output_directory=OUTPUT_DIRECTORY,
                      seed=SEED,
                      iteration_list=2 ** np.arange(10),
                      max_attempts=500,
                      keep_percent_list=[0.25, 0.5, 0.75])

    df_run_stats, df_run_curves = mmc.run()
"""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np
import pandas as pd

from mlrose_ky.algorithms import mimic
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase


@short_name("mimic")
class MIMICRunner(_RunnerBase):
    """
    A runner for performing optimization experiments using MIMIC.

    This class extends _RunnerBase and provides functionality for running experiments
    with the MIMIC algorithm, including grid search over hyperparameters such as
    population size and keep percent.

    Attributes
    ----------
    keep_percent_list : list[float]
        List of keep percentages to test in the grid search.
    population_sizes : list[int]
        List of population sizes to test in the grid search.
    _use_fast_mimic : bool, optional
        Whether to use the fast MIMIC mode, if available.
    """

    def __init__(
        self,
        problem: Any,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        population_sizes: list[int],
        keep_percent_list: list[float],
        max_attempts: int = 5,
        generate_curves: bool = True,
        use_fast_mimic: bool = True,
        output_directory: str = None,
        **kwargs: Any,
    ):
        """
        Initialize the MIMICRunner class with problem data and various experiment parameters.

        Parameters
        ----------
        problem : Any
            The optimization problem to be solved.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed for reproducibility.
        iteration_list : np.ndarray | list of int
            List of iterations for the experiment.
        population_sizes : list of int
            List of population sizes to test in the grid search.
        keep_percent_list : list of float
            List of keep percentages to test in the grid search.
        max_attempts : int, optional
            Maximum number of attempts without improvement before stopping.
        generate_curves : bool, optional
            Whether to generate learning curves.
        use_fast_mimic : bool, optional
            Whether to use the fast MIMIC mode, if available.
        output_directory : str, optional
            Directory to save experiment result, default=None.
        """
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            output_directory=output_directory,
            **kwargs,
        )
        self.keep_percent_list: list[float] = keep_percent_list
        self.population_sizes: list[int] = population_sizes
        self._use_fast_mimic: bool | None = None

        # Set fast MIMIC mode if available
        if hasattr(problem, "set_mimic_fast_mode") and callable(getattr(problem, "set_mimic_fast_mode")):
            self._use_fast_mimic = use_fast_mimic
            problem.set_mimic_fast_mode(use_fast_mimic)

    def _setup(self):
        """
        Perform any necessary setup before running the experiment.

        Logs the current state of the fast MIMIC mode if it is enabled.
        """
        super()._setup()

        if self._use_fast_mimic is not None:
            self._log_current_argument("use_fast_mimic", self._use_fast_mimic)

    def run(self) -> tuple[pd.DataFrame | None, pd.DataFrame | None]:
        """
        Run the MIMIC algorithm experiment.

        This method performs grid search over the provided population sizes
        and keep percentages and returns the statistics and curves generated by the experiment.

        Returns
        -------
        tuple
            A tuple containing two DataFrames: run statistics and run curves
        """
        return super().run_experiment_(
            algorithm=mimic, pop_size=("Population Size", self.population_sizes), keep_pct=("Keep Percent", self.keep_percent_list)
        )



"""Class defining a Four Peaks optimization problem generator."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

import numpy as np

from mlrose_ky import DiscreteOpt, FourPeaks


class FourPeaksGenerator:
    """A class to generate Four Peaks optimization problem instances."""

    @staticmethod
    def generate(seed: int = 42, size: int = 20, t_pct: float = 0.1) -> DiscreteOpt:
        """
        Generate a Four Peaks optimization problem with a given seed, size, and threshold percentage.

        Parameters
        ----------
        seed : int, optional, default=42
            The seed for the random number generator.
        size : int, optional, default=20
            The size of the problem.
        t_pct : float, optional, default=0.1
            The threshold percentage for the Four Peaks problem.

        Returns
        -------
        DiscreteOpt
            An instance of DiscreteOpt configured with the specified parameters.

        Raises
        ------
        ValueError
            If the size is not a positive integer or if the `t_pct` is not between 0 and 1.
        """
        if size <= 0:
            raise ValueError(f"Size must be a positive integer. Got {size}.")
        if not (0 <= t_pct <= 1):
            raise ValueError(f"Threshold percentage must be between 0 and 1. Got {t_pct}.")

        np.random.seed(seed)

        fitness = FourPeaks(t_pct=t_pct)
        return DiscreteOpt(length=size, fitness_fn=fitness)



"""Class for running optimization experiments using sklearn's MLPClassifier, including grid search functionality."""

# Authors: Andrew Rollings (modified by Kyle Nakamura)
# License: BSD 3-clause

import inspect
from typing import Any

import numpy as np
import sklearn.metrics as skmt
from sklearn.base import BaseEstimator
from sklearn.neural_network import MLPClassifier

import mlrose_ky.neural.activation as act
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._nn_runner_base import _NNRunnerBase


@short_name("skmlp")
class SKMLPRunner(_NNRunnerBase):
    """
    A runner for performing optimization experiments using sklearn's MLPClassifier.

    This class extends _NNRunnerBase and provides grid search functionality for optimizing
    the hyperparameters of an MLPClassifier.

    Attributes
    ----------
    classifier : SKMLPRunner._MLPClassifier
        An instance of an internal MLPClassifier with extended functionality.
    """

    class _MLPClassifier(BaseEstimator):
        """
        Internal wrapper for MLPClassifier with additional callback functionality for tracking
        the training progress and storing intermediate results.

        Parameters
        ----------
        runner
            Reference to the parent runner class.
        **kwargs : dict
            Additional keyword arguments to be passed to MLPClassifier.
        """

        def __init__(self, runner: "SKMLPRunner", **kwargs):
            """
            Initialize the _MLPClassifier wrapper.

            Parameters
            ----------
            runner : SKMLPRunner
                The parent runner instance.
            **kwargs : dict
                Additional arguments to pass to MLPClassifier.
            """
            self.runner: SKMLPRunner = runner

            # Filter out invalid kwargs using inspect.signature
            valid_args = inspect.signature(MLPClassifier.__init__).parameters
            mlp_kwargs = {k: v for k, v in kwargs.items() if k in valid_args}

            self.mlp: MLPClassifier = MLPClassifier(**mlp_kwargs)
            self.state_callback = self.runner._save_state
            self.fit_started_: bool = False
            self.user_info_: list[tuple[str, Any]] | None = None
            self.kwargs_: dict[str, Any] = kwargs
            self.loss_: float = 1.0
            self.state_: list | None = None
            self.curve_: list[tuple[float, int | None]] = []

            # need to intercept the classifier so we can track statistics.
            if runner.generate_curves:
                if hasattr(self.mlp, "_update_no_improvement_count"):
                    # noinspection PyProtectedMember
                    self._mlp_update_no_improvement_count = self.mlp._update_no_improvement_count
                    self.mlp._update_no_improvement_count = self._update_no_improvement_count_intercept
                if hasattr(self.mlp, "_loss_grad_lbfgs"):
                    # noinspection PyProtectedMember
                    self._mlp_loss_grad_lbfgs = self.mlp._loss_grad_lbfgs
                    self.mlp._loss_grad_lbfgs = self._loss_grad_lbfgs_intercept

        def __getattr__(self, item: str, default: Any = None) -> Any:
            """Fetch an attribute from the MLPClassifier if not found in the class instance."""
            if "mlp" in self.__dict__ and hasattr(self.__dict__["mlp"], item):
                return self.__dict__["mlp"].__getattr__(item, default)

            return self.__dict__[item] if item in self.__dict__ else default

        def __setattr__(self, item: str, value: Any):
            """Set an attribute on the MLPClassifier if it exists, otherwise set it on the class instance."""
            if "mlp" in self.__dict__ and hasattr(self.__dict__["mlp"], item):
                self.__dict__["mlp"].__setattr__(item, value)

            self.__dict__[item] = value

        def get_params(self, deep: bool = True) -> dict:
            """
            Get parameters of the MLPClassifier and this wrapper.

            Parameters
            ----------
            deep : bool, optional
                If True, will return the parameters for this estimator and contained sub-objects that are estimators.

            Returns
            -------
            dict
                A dictionary of parameters.
            """
            out = super().get_params()
            out.update(self.mlp.get_params())

            # Exclude any that end with an underscore
            return {k: v for (k, v) in out.items() if not k[-1] == "_"}

        def fit(self, x_train: np.ndarray, y_train: np.ndarray = None) -> MLPClassifier:
            """
            Fit the model to the training data.

            Parameters
            ----------
            x_train : np.ndarray
                Training input data.
            y_train : np.ndarray, optional
                Target labels for training data.

            Returns
            -------
            MLPClassifier
                The trained model.
            """
            self.fit_started_ = True
            self.runner._start_run_timing()

            # Make initial callback
            self._invoke_runner_callback()

            # noinspection PyTypeChecker
            return self.mlp.fit(x_train, y_train)

        def predict(self, x_test: np.ndarray) -> np.ndarray:
            """
            Predict the class labels for the provided data.

            Parameters
            ----------
            x_test : {array-like, sparse matrix} of shape (n_samples, n_features)
                The input data.

            Returns
            -------
            y : ndarray, shape (n_samples,) or (n_samples, n_classes)
                The predicted classes.
            """
            return self.mlp.predict(x_test)

        def _update_no_improvement_count_intercept(self, early_stopping: bool, x_val: np.ndarray, y_val: np.ndarray) -> int:
            """
            Intercept the '_update_no_improvement_count' method to track state.

            Parameters
            ----------
            early_stopping : bool
                Whether early stopping is enabled.
            x_val : np.ndarray
                Validation input data.
            y_val : np.ndarray
                Validation target data.

            Returns
            -------
            int
                The result of the original '_update_no_improvement_count' method.
            """
            ret = self._mlp_update_no_improvement_count(early_stopping, x_val, y_val)

            self.state_ = self.mlp.coefs_ if hasattr(self.mlp, "coefs_") else []
            self.loss_ = self.mlp.loss_ if hasattr(self.mlp, "loss_") else 0

            if hasattr(self.mlp, "loss_curve_"):
                self.curve_ = [(_loss_val, None) for _loss_val in self.mlp.loss_curve_]
            else:
                self.curve_.append((self.loss_, None))

            self._invoke_runner_callback()

            return ret

        def _loss_grad_lbfgs_intercept(
            self,
            packed_coef_inter: np.ndarray,
            x: np.ndarray,
            y: np.ndarray,
            activations: list,
            deltas: list,
            coef_grads: list,
            intercept_grads: list,
        ) -> tuple[float, np.ndarray | list]:
            """
            Intercept the '_loss_grad_lbfgs' method to track loss and state.

            Parameters
            ----------
            packed_coef_inter : ndarray
                A vector comprising the flattened coefficients and intercepts.

            x : {array-like, sparse matrix} of shape (n_samples, n_features)
                The input data.

            y : ndarray of shape (n_samples,)
                The target values.

            activations : list, length = n_layers - 1
                The ith element of the list holds the values of the ith layer.

            deltas : list, length = n_layers - 1
                The ith element of the list holds the difference between the
                activations of the i + 1 layer and the backpropagated error.
                More specifically, deltas are gradients of loss with respect to z
                in each layer, where z = wx + b is the value of a particular layer
                before passing through the activation function

            coef_grads : list, length = n_layers - 1
                The ith element contains the amount of change used to update the
                coefficient parameters of the ith layer in an iteration.

            intercept_grads : list, length = n_layers - 1
                The ith element contains the amount of change used to update the
                intercept parameters of the ith layer in an iteration.

            Returns
            -------
            loss : float
            grad : array-like, shape (number of nodes of all layers,)
            """
            f, g = self._mlp_loss_grad_lbfgs(packed_coef_inter, x, y, activations, deltas, coef_grads, intercept_grads)

            self.loss_ = f
            self.state_ = g
            self.curve_.append((self.loss_, None))
            self._invoke_runner_callback()

            return f, g

        def _invoke_runner_callback(self):
            """Invoke the runner callback to save the current state of training."""
            # noinspection PyProtectedMember
            no_improvement_count = self.mlp._no_improvement_count if hasattr(self.mlp, "_no_improvement_count") else 0

            iterations = self.mlp.n_iter_ if hasattr(self.mlp, "n_iter_") else 0
            done = self.mlp.early_stopping and (no_improvement_count > self.mlp.n_iter_no_change) or iterations == self.mlp.max_iter

            # Check for early abort
            if self.runner.has_aborted():
                return

            if self.user_info_ is None:
                self.user_info_ = [(k, self.__dict__[k]) for k in self.kwargs_.keys() if hasattr(self, k)]
                for k, v in self.user_info_:
                    self.runner._log_current_argument(k, v)

            self.state_callback(
                iteration=iterations,
                state=self.state_,
                fitness=self.loss_,
                user_data=self.user_info_,
                attempt=no_improvement_count,
                done=done,
                curve=self.curve_,
            )

    def __init__(
        self,
        x_train: np.ndarray,
        y_train: np.ndarray,
        x_test: np.ndarray,
        y_test: np.ndarray,
        experiment_name: str,
        seed: int,
        iteration_list: np.ndarray | list[int],
        grid_search_parameters: dict[str, Any],
        grid_search_scorer_method: callable = skmt.balanced_accuracy_score,
        early_stopping: bool = True,
        max_attempts: int = 500,
        n_jobs: int = 1,
        cv: int = 5,
        override_ctrl_c_handler: bool = True,
        generate_curves: bool = True,
        output_directory: str = None,
        replay: bool = False,
        **kwargs: dict,
    ):
        """
        Initialize the SKMLPRunner class with training and testing data and various experiment parameters.

        Parameters
        ----------
        x_train : np.ndarray
            Training input data.
        y_train : np.ndarray
            Target labels for training data.
        x_test : np.ndarray
            Test input data.
        y_test : np.ndarray
            Target labels for test data.
        experiment_name : str
            Name of the experiment.
        seed : int
            Random seed for reproducibility.
        iteration_list : np.ndarray | list of int
            List of iterations for the experiment.
        grid_search_parameters : dict
            Parameters for grid search.
        grid_search_scorer_method : callable, optional
            Scoring method for grid search.
        early_stopping : bool, optional
            Whether to stop early if no improvement is detected.
        max_attempts : int, optional
            Maximum number of attempts without improvement before stopping.
        n_jobs : int, optional
            Number of jobs to run in parallel.
        cv : int, optional
            Number of cross-validation folds.
        override_ctrl_c_handler : bool, optional
            Whether to override the Ctrl-C handler.
        generate_curves : bool, optional
            Whether to generate learning curves.
        output_directory : str, optional
            Directory to save output.
        replay : bool, optional
            Whether to replay the experiment.
        """
        grid_search_parameters = {**grid_search_parameters}

        # Hack for compatibility purposes
        if "max_iters" in grid_search_parameters:
            grid_search_parameters["max_iter"] = grid_search_parameters.pop("max_iters")

        if "max_attempts" in grid_search_parameters:
            grid_search_parameters["n_iter_no_change"] = grid_search_parameters.pop("max_attempts")

        super().__init__(
            x_train=x_train,
            y_train=y_train,
            x_test=x_test,
            y_test=y_test,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            grid_search_parameters=grid_search_parameters,
            grid_search_scorer_method=grid_search_scorer_method,
            generate_curves=generate_curves,
            output_directory=output_directory,
            override_ctrl_c_handler=override_ctrl_c_handler,
            replay=replay,
            n_jobs=n_jobs,
            cv=cv,
        )

        # Create a dictionary of default values
        default_kwargs = {
            "shuffle": True,
            "random_state": seed,
            "verbose": False,
            "warm_start": False,
            "early_stopping": early_stopping,
            "n_iter_no_change": max_attempts,
        }

        # Merge default_kwargs into kwargs, with kwargs taking precedence in case of conflicts
        kwargs = {**default_kwargs, **kwargs}

        # Instantiate the _MLPClassifier with runner and kwargs
        self.classifier = self._MLPClassifier(runner=self, **kwargs)  # Pass everything as kwargs

        self.classifier.runner = self

    @staticmethod
    def build_grid_search_parameters(grid_search_parameters: dict[str, Any], **kwargs: dict) -> dict[str, Any]:
        """
        Build and return grid search parameters, ensuring compatibility with sklearn's MLPClassifier.

        Parameters
        ----------
        grid_search_parameters : dict
            Initial grid search parameters.
        **kwargs : dict
            Additional parameters.

        Returns
        -------
        dict
            Final grid search parameters with compatible types.
        """
        all_grid_search_parameters = _NNRunnerBase.build_grid_search_parameters(grid_search_parameters, **kwargs)

        if "activation" in all_grid_search_parameters:
            activation_set = list(all_grid_search_parameters["activation"])

            for i in range(len(activation_set)):
                a = activation_set[i]
                if a == act.relu:
                    activation_set[i] = "relu"
                elif a == act.sigmoid:
                    activation_set[i] = "logistic"
                elif a == act.tanh:
                    activation_set[i] = "tanh"
                elif a == act.identity:
                    activation_set[i] = "identity"
                elif a == act.softmax:
                    activation_set[i] = "softmax"

            all_grid_search_parameters["activation"] = activation_set

        return all_grid_search_parameters



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from typing import Any

import numpy as np

from mlrose_ky.algorithms.crossovers import TSPCrossOver
from mlrose_ky.algorithms.mutators import SwapMutator
from mlrose_ky.fitness import TravellingSales
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class TSPOpt(DiscreteOpt):
    """Class for defining travelling salesperson optimization problems.

    Parameters
    ----------
    length : int, default=None
        Number of elements in state vector. Must equal the number of nodes in the tour.

    fitness_fn : TravellingSalesperson, default=None
        Object to implement the fitness function for optimization.
        If None, a `TravellingSalesperson` fitness function is created using `coords` or `distances`.

    maximize : bool, default=False
        Whether to maximize the fitness function. Set :code:`False` for minimization problems.

    coords : list[tuple], default=None
        Ordered list of the (x, y) coordinates of all nodes. Used to calculate distances if no `fitness_fn` is provided.

    distances : list[tuple[int, int, float]], default=None
        List giving the distances between pairs of nodes for which travel is possible.
        Each list item is in the form (u, v, d), representing nodes u and v, and the distance d.

    source_graph : networkx.Graph, default=None
        Optional graph object for representing the TSP problem structure.
    """

    def __init__(
        self,
        length: int = None,
        fitness_fn: Any = None,
        maximize: bool = False,
        coords: list[tuple] = None,
        distances: list[tuple] = None,
        source_graph: Any = None,
    ):
        # Ensure that at least one of fitness_fn, coords, or distances is provided
        if fitness_fn is None and coords is None and distances is None:
            raise ValueError("At least one of fitness_fn, coords, or distances must be specified.")

        # If fitness_fn is not provided, create a TravellingSalesperson fitness function
        if fitness_fn is None:
            fitness_fn = TravellingSales(coords=coords, distances=distances)

        self.distances: list[tuple] | None = distances
        self.coords: list[tuple] | None = coords

        # If length is not provided, infer it from coords or distances
        if length is None:
            if coords is not None:
                length = len(coords)
            elif distances is not None:
                # Calculate length from the set of nodes in the distances list
                length = len(set([x for (x, _, _) in distances] + [x for (_, x, _) in distances]))

        self.length: int = length

        # Initialize the parent class with the TSPCrossOver and SwapMutator
        super().__init__(length, fitness_fn, maximize, max_val=length, crossover=TSPCrossOver(self), mutator=SwapMutator(self))

        # Ensure that the fitness function type is 'tsp'
        if self.fitness_fn.get_prob_type() != "tsp":
            raise ValueError("fitness_fn must have problem type 'tsp'.")

        self.source_graph = source_graph
        self.prob_type = "tsp"

    @staticmethod
    def adjust_probs(probs: np.ndarray) -> np.ndarray:
        """Normalize a vector of probabilities so that the vector sums to 1.

        Parameters
        ----------
        probs : np.ndarray
            Vector of probabilities that may or may not sum to 1.

        Returns
        -------
        adj_probs : np.ndarray
            Vector of probabilities that sums to 1. Returns a zero vector if sum(probs) = 0.
        """
        sp = np.sum(probs)
        return np.zeros(np.shape(probs)) if sp == 0 else probs / sp

    def find_neighbors(self):
        """Find all neighbors of the current state."""
        self.neighbors = []

        for node1 in range(self.length - 1):
            for node2 in range(node1 + 1, self.length):
                neighbor = np.copy(self.state)

                # Swap the positions of node1 and node2
                neighbor[node1] = self.state[node2]
                neighbor[node2] = self.state[node1]
                self.neighbors.append(neighbor)

    def random(self) -> np.ndarray:
        """Return a random state vector.

        Returns
        -------
        np.ndarray
            Randomly generated state vector (a random permutation of nodes).
        """
        return np.random.permutation(self.length)

    def random_mimic(self) -> np.ndarray:
        """Generate single MIMIC sample from probability density.

        Returns
        -------
        np.ndarray
            State vector of MIMIC random sample.
        """
        remaining = list(np.arange(self.length))
        state = np.zeros(self.length, dtype=np.int8)
        node_probs = np.copy(self.node_probs)

        # Get value of the first element in the new sample
        state[0] = np.random.choice(self.length, p=node_probs[0, 0])
        remaining.remove(state[0])
        node_probs[:, :, state[0]] = 0

        # Get the sample order
        self.find_sample_order()
        sample_order = self.sample_order[1:]

        # Set values of remaining elements of state
        for i in sample_order:
            par_ind = self.parent_nodes[i - 1]
            par_value = state[par_ind]
            probs = node_probs[i, par_value]

            if np.sum(probs) == 0:
                next_node = np.random.choice(remaining)
            else:
                adj_probs = self.adjust_probs(probs)
                next_node = np.random.choice(self.length, p=adj_probs)

            state[i] = next_node
            remaining.remove(next_node)
            node_probs[:, :, next_node] = 0

        return state

    def random_neighbor(self) -> np.ndarray:
        """Return random neighbor of current state vector.

        Returns
        -------
        neighbor : np.ndarray
            State vector of random neighbor.
        """
        neighbor = np.copy(self.state)
        node1, node2 = np.random.choice(np.arange(self.length), size=2, replace=False)

        # Swap the positions of node1 and node2
        neighbor[node1] = self.state[node2]
        neighbor[node2] = self.state[node1]

        return neighbor

    def sample_pop(self, sample_size: int) -> np.ndarray:
        """Generate a new sample from the probability density.

        Parameters
        ----------
        sample_size : int
            Size of the sample to be generated.

        Returns
        -------
        new_sample : np.ndarray
            Numpy array containing the new sample.
        """
        if sample_size <= 0:
            raise ValueError("sample_size must be a positive integer.")
        elif not isinstance(sample_size, int):
            if sample_size.is_integer():
                sample_size = int(sample_size)
            else:
                raise ValueError("sample_size must be a positive integer.")

        self.find_sample_order()
        new_sample = []

        # Generate MIMIC samples
        for _ in range(sample_size):
            state = self.random_mimic()
            new_sample.append(state)

        return np.array(new_sample)



"""Classes for defining mutation strategies for Genetic Algorithms (GA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

from .change_one_mutator import ChangeOneMutator
from .discrete_mutator import DiscreteMutator
from .shift_one_mutator import ShiftOneMutator
from .swap_mutator import SwapMutator



"""Class defining the Discrete Peaks base fitness function for use with the Four Peaks, Six Peaks, and Custom fitness functions."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3-clause

import numpy as np


class _DiscretePeaksBase:
    """Base class for Discrete Peaks fitness functions.

    This class provides methods to determine the number of leading or trailing
    specific integer values within a vector.
    """

    @staticmethod
    def head(_b: int, _x: np.ndarray) -> int:
        """Determine the number of leading occurrences of `_b` in vector `_x`.

        Parameters
        ----------
        _b : int
            The integer value to count at the beginning of the vector.
        _x : np.ndarray
            A vector of integers.

        Returns
        -------
        int
            Number of leading occurrences of `_b` in `_x`.

        Raises
        ------
        TypeError
            If `_x` is not an instance of `np.ndarray`.
        """
        if not isinstance(_x, np.ndarray):
            raise TypeError(f"Expected vector to be np.ndarray, got {type(_x).__name__} instead.")

        # Use NumPy's cumulative sum to find leading values
        leading_mask = np.cumsum(_x != _b) == 0
        return np.sum(leading_mask)

    @staticmethod
    def tail(_b: int, _x: np.ndarray) -> int:
        """Determine the number of trailing occurrences of `_b` in vector `_x`.

        Parameters
        ----------
        _b : int
            The integer value to count at the end of the vector.
        _x : np.ndarray
            A vector of integers.

        Returns
        -------
        int
            Number of trailing occurrences of `_b` in `_x`.

        Raises
        ------
        TypeError
            If `_x` is not an instance of `np.ndarray`.
        """
        if not isinstance(_x, np.ndarray):
            raise TypeError(f"Expected vector to be np.ndarray, got {type(_x).__name__} instead.")

        # Use NumPy's cumulative sum to find trailing values
        trailing_mask = np.cumsum(_x[::-1] != _b) == 0
        return np.sum(trailing_mask)
